# Copyright (c) 2010-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    TYPE_CHECKING,
    TypeVar,
)
from copy import deepcopy

from ezdxf.lldxf import const
from ezdxf.enums import (
    MTextEntityAlignment,
    MAP_MTEXT_ALIGN_TO_FLAGS,
)
from ezdxf.addons import MTextSurrogate
from ezdxf.math import UVec, Vec2

if TYPE_CHECKING:
    from ezdxf.layouts import BlockLayout
    from ezdxf.eztypes import GenericLayoutType

DEFAULT_TABLE_BG_LAYER = "TABLEBACKGROUND"
DEFAULT_TABLE_FG_LAYER = "TABLECONTENT"
DEFAULT_TABLE_GRID_LAYER = "TABLEGRID"
DEFAULT_TEXT_STYLE = "STANDARD"
DEFAULT_CELL_HEIGHT = 1.0
DEFAULT_CELL_WIDTH = 2.5
DEFAULT_CELL_CHAR_HEIGHT = 0.7
DEFAULT_CELL_LINE_SPACING = 1.5
DEFAULT_CELL_X_SCALE = 1.0
DEFAULT_CELL_Y_SCALE = 1.0
DEFAULT_CELL_TEXTCOLOR = const.BYLAYER
DEFAULT_CELL_BG_COLOR = None
DEFAULT_CELL_X_MARGIN = 0.1
DEFAULT_CELL_Y_MARGIN = 0.1
DEFAULT_BORDER_COLOR = 5
DEFAULT_BORDER_LINETYPE = "BYLAYER"
DEFAULT_BORDER_STATUS = True
DEFAULT_BORDER_PRIORITY = 50


T = TypeVar("T", bound="Cell")


class TablePainter:
    """The TablePainter class renders tables build from DXF primitives.

    The TablePainter instance contains all the data cells.

    Args:
        insert: insert location as or :class:`~ezdxf.math.UVec`
        nrows: row count
        ncols: column count
        cell_width: default cell width in drawing units
        cell_height: default cell height in drawing units
        default_grid: draw a grid of solid lines if ``True``, otherwise
            draw only explicit defined borders, the default grid has a
            priority of 50.

    """

    def __init__(
        self,
        insert: UVec,
        nrows: int,
        ncols: int,
        cell_width=DEFAULT_CELL_WIDTH,
        cell_height=DEFAULT_CELL_HEIGHT,
        default_grid=True,
    ):
        self.insert = Vec2(insert)
        self.nrows: int = nrows
        self.ncols: int = ncols
        self.row_heights: list[float] = [cell_height] * nrows
        self.col_widths: list[float] = [cell_width] * ncols
        self.bg_layer_name: str = DEFAULT_TABLE_BG_LAYER
        self.fg_layer_name: str = DEFAULT_TABLE_FG_LAYER
        self.grid_layer_name: str = DEFAULT_TABLE_GRID_LAYER
        self.styles: dict[str, CellStyle] = {"default": CellStyle()}
        if not default_grid:
            default_style = self.get_cell_style("default")
            default_style.set_border_status(False, False, False, False)

        self._cells: dict[tuple[int, int], Cell] = {}  # data cells
        self.frames: list[Frame] = []  # border frame objects
        self.empty_cell = Cell(self)  # represents all empty cells

    def set_col_width(self, index: int, value: float):
        """Set column width in drawing units of the given column index.

        Args:
            index: zero based column index
            value: new column width in drawing units

        """
        self.col_widths[index] = float(value)

    def set_row_height(self, index: int, value: float):
        """Set row height in drawing units of the given row index.

        Args:
            index: zero based row index
            value: new row height in drawing units
        """
        self.row_heights[index] = float(value)

    def text_cell(
        self,
        row: int,
        col: int,
        text: str,
        span: tuple[int, int] = (1, 1),
        style="default",
    ) -> TextCell:
        """Factory method to create a new text cell at location (row, col),
        with `text` as content, the `text` can be a line breaks ``'\\n'``.
        The final cell can spread over several cells defined by the argument
        `span`.

        """
        cell = TextCell(self, text, style=style, span=span)
        return self.set_cell(row, col, cell)

    def block_cell(
        self,
        row: int,
        col: int,
        blockdef: BlockLayout,
        span: tuple[int, int] = (1, 1),
        attribs=None,
        style="default",
    ) -> BlockCell:
        """Factory method to Create a new block cell at position (row, col).

        Content is a block reference inserted by an INSERT entity,
        attributes will be added if the block definition contains ATTDEF.
        Assignments are defined by attribs-key to attdef-tag association.

        Example: attribs = {'num': 1} if an ATTDEF with tag=='num' in
        the block definition exists, an attrib with text=str(1) will be
        created and added to the insert entity.

        The cell spans over 'span' cells and has the cell style with the
        name 'style'.
        """
        if attribs is None:
            attribs = {}
        cell = BlockCell(
            self, blockdef, style=style, attribs=attribs, span=span
        )
        return self.set_cell(row, col, cell)

    @property
    def table_width(self) -> float:
        """Returns the total table width."""
        return sum(self.col_widths)

    @property
    def table_height(self) -> float:
        """Returns the total table height."""
        return sum(self.row_heights)

    def set_cell(self, row: int, col: int, cell: T) -> T:
        """Insert a cell at position (row, col)."""
        row, col = self.validate_index(row, col)
        self._cells[row, col] = cell
        return cell

    def get_cell(self, row: int, col: int) -> Cell:
        """Get cell at location (row, col)."""
        row, col = self.validate_index(row, col)
        try:
            return self._cells[row, col]
        except KeyError:
            return self.empty_cell  # empty cell with default style

    def validate_index(self, row: int, col: int) -> tuple[int, int]:
        row = int(row)
        col = int(col)
        if row < 0 or row >= self.nrows or col < 0 or col >= self.ncols:
            raise IndexError("cell index out of range")
        return row, col

    def frame(
        self,
        row: int,
        col: int,
        width: int = 1,
        height: int = 1,
        style="default",
    ) -> Frame:
        """Creates a frame around the give cell area, starting at (row, col) and
        covering `width` columns and `height` rows. The `style` argument is the
        name of a :class:`CellStyle`.
        """
        frame = Frame(self, pos=(row, col), span=(height, width), style=style)
        self.frames.append(frame)
        return frame

    def new_cell_style(self, name: str, **kwargs) -> CellStyle:
        """Factory method to create a new :class:`CellStyle` object, overwrites
        an already existing cell style.

        Args:
            name: style name as string
            kwargs: see attributes of class :class:`CellStyle`

        """
        assert (
            isinstance(name, str) and name != ""
        ), "name has to be a non-empty string"
        style: CellStyle = deepcopy(self.get_cell_style("default"))
        style.update(kwargs)
        self.styles[name] = style
        return style

    @staticmethod
    def new_border_style(
        color: int = const.BYLAYER,
        status=True,
        priority: int = 100,
        linetype: str = "BYLAYER",
        lineweight: int = const.LINEWEIGHT_BYLAYER,
    ) -> BorderStyle:
        """Factory method to create a new border style.

        Args:
            status: ``True`` for visible, ``False`` for invisible
            color: :ref:`ACI`
            linetype: linetype name, default is "BYLAYER"
            lineweight: lineweight as int, default is by layer
            priority: drawing priority, higher priorities cover lower priorities

        """
        border_style = BorderStyle()
        border_style.color = color
        border_style.linetype = linetype
        border_style.lineweight = lineweight
        border_style.status = status
        border_style.priority = priority
        return border_style

    def get_cell_style(self, name: str) -> CellStyle:
        """Get cell style by name."""
        return self.styles[name]

    def iter_visible_cells(
        self, visibility_map: VisibilityMap
    ) -> Iterator[tuple[int, int, Cell]]:
        """Iterate over all visible cells"""
        return (
            (row, col, self.get_cell(row, col)) for row, col in visibility_map
        )

    def render(self, layout: GenericLayoutType, insert: Optional[UVec] = None):
        """Render table to layout."""
        insert_backup = self.insert
        if insert is not None:
            self.insert = Vec2(insert)
        visibility_map = VisibilityMap(self)
        grid = Grid(self)
        for row, col, cell in self.iter_visible_cells(visibility_map):
            grid.render_cell_background(layout, row, col, cell)
            grid.render_cell_content(layout, row, col, cell)
        grid.render_lines(layout, visibility_map)
        self.insert = insert_backup


class VisibilityMap:
    """Stores the visibility of the table cells."""

    def __init__(self, table: TablePainter):
        """Create the visibility map for table."""
        self.table: TablePainter = table
        self._hidden_cells: set[tuple[int, int]] = set()
        self._create_visibility_map()

    def _create_visibility_map(self):
        """Set visibility for all existing cells."""
        for row, col in iter(self):
            cell = self.table.get_cell(row, col)
            self._set_span_visibility(row, col, cell.span)

    def _set_span_visibility(self, row: int, col: int, span: tuple[int, int]):
        """Set the visibility of the given cell.

        The cell itself is visible, all other cells in the span-range
        (tuple: width, height) are invisible, they are covered by the
        main cell (row, col).
        """

        if span != (1, 1):
            nrows, ncols = span
            for rowx in range(nrows):
                for colx in range(ncols):
                    # switch all cells in span range to invisible
                    self.hide(row + rowx, col + colx)
        # switch content cell visible
        self.show(row, col)

    def show(self, row: int, col: int):
        """Show cell (row, col)."""
        try:
            self._hidden_cells.remove((row, col))
        except KeyError:
            pass

    def hide(self, row: int, col: int) -> None:
        """Hide cell (row, col)."""
        self._hidden_cells.add((row, col))

    def iter_all_cells(self) -> Iterator[tuple[int, int]]:
        """Iterate over all cell indices, yields (row, col) tuples."""
        for row in range(self.table.nrows):
            for col in range(self.table.ncols):
                yield row, col

    def is_visible_cell(self, row: int, col: int) -> bool:
        """True if cell (row, col)  is visible, else False."""
        return (row, col) not in self._hidden_cells

    def __iter__(self) -> Iterator[tuple[int, int]]:
        """Iterate over all visible cells."""
        return (
            (row, col)
            for (row, col) in self.iter_all_cells()
            if self.is_visible_cell(row, col)
        )


class CellStyle:
    """Cell style object.

    .. important::

        Always instantiate new styles by the factory method:
        :meth:`TablePainter.new_cell_style`

    """

    def __init__(self, data: Optional[dict[str, Any]] = None):
        # text style is ignored by block cells
        self.text_style = "STANDARD"
        # text height in drawing units, ignored by block cells
        self.char_height = DEFAULT_CELL_CHAR_HEIGHT
        # line spacing in percent = char_height * line_spacing, ignored by block cells
        self.line_spacing = DEFAULT_CELL_LINE_SPACING
        # text stretching factor (width factor) or block reference x-scaling factor
        self.scale_x = DEFAULT_CELL_X_SCALE
        # block reference y-axis scaling factor, ignored by text cells
        self.scale_y = DEFAULT_CELL_Y_SCALE
        # dxf color index, ignored by block cells
        self.text_color = DEFAULT_CELL_TEXTCOLOR
        # text or block rotation in degrees
        self.rotation = 0.0
        # Letters are stacked top-to-bottom, but not rotated
        self.stacked = False
        # text and block alignment, see ezdxf.enums.MTextEntityAlignment
        self.align = MTextEntityAlignment.TOP_CENTER
        # left and right cell margin in drawing units
        self.margin_x = DEFAULT_CELL_X_MARGIN
        # top and bottom cell margin in drawing units
        self.margin_y = DEFAULT_CELL_Y_MARGIN
        # background color, dxf color index, ignored by block cells
        self.bg_color = DEFAULT_CELL_BG_COLOR
        # left border style
        self.left = BorderStyle()
        # top border style
        self.top = BorderStyle()
        # right border style
        self.right = BorderStyle()
        # bottom border style
        self.bottom = BorderStyle()
        if data:
            self.update(data)

    def __getitem__(self, k: str) -> Any:
        return self.__dict__[k]

    def __setitem__(self, k: str, v: Any):
        if k in self.__dict__:
            self.__dict__.__setitem__(k, v)
        else:
            raise KeyError(f"invalid attribute name: {k}")

    def update(self, data: dict[str, Any]):
        for k, v in data.items():
            self.__setitem__(k, v)
        assert isinstance(
            self.align, MTextEntityAlignment
        ), "enum ezdxf.enums.MTextEntityAlignment for text alignments required"

    def set_border_status(self, left=True, right=True, top=True, bottom=True):
        """Set status of all cell borders at once."""
        self.left.status = left
        self.right.status = right
        self.top.status = top
        self.bottom.status = bottom

    def set_border_style(
        self, style: BorderStyle, left=True, right=True, top=True, bottom=True
    ):
        """Set border styles of all cell borders at once."""
        for border, status in (
            ("left", left),
            ("right", right),
            ("top", top),
            ("bottom", bottom),
        ):
            if status:
                self[border] = style

    @staticmethod
    def get_default_border_style() -> BorderStyle:
        return BorderStyle()

    def get_text_align_flags(self) -> tuple[int, int]:
        return MAP_MTEXT_ALIGN_TO_FLAGS[self.align]


class BorderStyle:
    """Border style class.

    .. important::

        Always instantiate new border styles by the factory method:
        :meth:`TablePainter.new_border_style`

    """

    def __init__(
        self,
        status: bool = DEFAULT_BORDER_STATUS,
        color: int = DEFAULT_BORDER_COLOR,
        linetype: str = DEFAULT_BORDER_LINETYPE,
        lineweight=const.LINEWEIGHT_BYLAYER,
        priority: int = DEFAULT_BORDER_PRIORITY,
    ):
        # border status, True for visible, False for hidden
        self.status = status
        # ACI
        self.color = color
        # linetype name, BYLAYER if None
        self.linetype = linetype
        # lineweight
        self.lineweight = lineweight
        # drawing priority, higher values cover lower values
        self.priority = priority


class Grid:
    """Grid contains the graphical representation of the table."""

    def __init__(self, table: TablePainter):
        self.table: TablePainter = table
        # contains the x-axis coords of the grid lines between the data columns.
        self.col_pos: list[float] = self._calc_col_pos()
        # contains the y-axis coords of the grid lines between the data rows.
        self.row_pos: list[float] = self._calc_row_pos()

        # _x_borders contains the horizontal border elements, list of border styles
        # get index with _border_index(row, col), which means the border element
        # above row, col, and row-indices are [0 .. nrows+1], nrows+1 for the
        # grid line below the last row; list contains only the border style with
        # the highest priority.
        self._x_borders: list[BorderStyle] = []  # created in _init_borders

        # _y_borders: same as _x_borders but for the vertical borders,
        # col-indices are [0 .. ncols+1], ncols+1 for the last grid line right
        # of the last column
        self._y_borders: list[BorderStyle] = []  # created in _init_borders
        # border style to delete borders inside of merged cells
        self.no_border = BorderStyle(
            status=False, priority=999, linetype="BYLAYER", color=0
        )

    def _init_borders(self, x_border: BorderStyle, y_border: BorderStyle):
        """Init the _hborders with  <hborder> and _vborders with <vborder>."""
        # <border_count> has more elements than necessary, but it unifies the
        # index calculation for _vborders and _hborders.
        # exact values are:
        # x_border_count = ncols * (nrows+1), hindex = ncols * <row> + <col>
        # y_border_count = nrows * (ncols+1), vindex = (ncols+1) * <row> + <col>
        border_count: int = (self.table.nrows + 1) * (self.table.ncols + 1)
        self._x_borders = [x_border] * border_count
        self._y_borders = [y_border] * border_count

    def _border_index(self, row: int, col: int) -> int:
        """Calculate linear index for border arrays _x_borders and _y_borders."""
        return row * (self.table.ncols + 1) + col

    def set_x_border(self, row: int, col: int, border_style: BorderStyle):
        """Set <border_style> for the horizontal border element above
        <row>, <col>.
        """
        return self._set_border_style(self._x_borders, row, col, border_style)

    def set_y_border(self, row: int, col: int, border_style: BorderStyle):
        """Set <border_style> for the vertical border element left of
        <row>, <col>.
        """
        return self._set_border_style(self._y_borders, row, col, border_style)

    def _set_border_style(
        self,
        borders: list[BorderStyle],
        row: int,
        col: int,
        border_style: BorderStyle,
    ):
        """Set <border_style> for <row>, <col> in <borders>."""
        border_index = self._border_index(row, col)
        actual_borderstyle = borders[border_index]
        if border_style.priority >= actual_borderstyle.priority:
            borders[border_index] = border_style

    def get_x_border(self, row: int, col: int) -> BorderStyle:
        """Get the horizontal border element above <row>, <col>.
        Last grid line (below <nrows>) is the element above of <nrows+1>.
        """
        return self._get_border(self._x_borders, row, col)

    def get_y_border(self, row: int, col: int) -> BorderStyle:
        """Get the vertical border element left of <row>, <col>.
        Last grid line (right of <ncols>) is the element left of <ncols+1>.
        """
        return self._get_border(self._y_borders, row, col)

    def _get_border(
        self, borders: list[BorderStyle], row: int, col: int
    ) -> BorderStyle:
        """Get border element at <row>, <col> from <borders>."""
        return borders[self._border_index(row, col)]

    def _calc_col_pos(self) -> list[float]:
        """Calculate the x-axis coords of the grid lines between the columns."""
        col_pos: list[float] = []
        start_x: float = self.table.insert.x
        sum_fields(start_x, self.table.col_widths, col_pos.append)
        return col_pos

    def _calc_row_pos(self) -> list[float]:
        """Calculate the y-axis coords of the grid lines between the rows."""
        row_pos: list[float] = []
        start_y: float = self.table.insert.y
        sum_fields(start_y, self.table.row_heights, row_pos.append, -1.0)
        return row_pos

    def cell_coords(
        self, row: int, col: int, span: tuple[int, int]
    ) -> tuple[float, float, float, float]:
        """Get the coordinates of the cell <row>,<col> as absolute drawing units.

        :return: a tuple (left, right, top, bottom)
        """
        top = self.row_pos[row]
        bottom = self.row_pos[row + span[0]]
        left = self.col_pos[col]
        right = self.col_pos[col + span[1]]
        return left, right, top, bottom

    def render_cell_background(
        self, layout: GenericLayoutType, row: int, col: int, cell: Cell
    ):
        """Render the cell background for (row, col) as SOLID entity."""
        style = cell.style
        if style.bg_color is None:
            return
        # get cell coords in absolute drawing units
        left, right, top, bottom = self.cell_coords(row, col, cell.span)
        layout.add_solid(
            points=((left, top), (left, bottom), (right, top), (right, bottom)),
            dxfattribs={
                "color": style.bg_color,
                "layer": self.table.bg_layer_name,
            },
        )

    def render_cell_content(
        self, layout: GenericLayoutType, row: int, col: int, cell: Cell
    ):
        """Render the cell content for <row>,<col> into layout object."""
        # get cell coords in absolute drawing units
        coords = self.cell_coords(row, col, cell.span)
        cell.render(layout, coords, self.table.fg_layer_name)

    def render_lines(self, layout: GenericLayoutType, vm: VisibilityMap):
        """Render all grid lines into layout object."""
        # Init borders with default_style top- and left border.
        default_style = self.table.get_cell_style("default")
        x_border = default_style.top
        y_border = default_style.left
        self._init_borders(x_border, y_border)
        self._set_frames(self.table.frames)
        self._set_borders(self.table.iter_visible_cells(vm))
        self._render_borders(layout, self.table)

    def _set_borders(self, visible_cells: Iterable[tuple[int, int, Cell]]):
        """Set borders of the visible cells."""
        for row, col, cell in visible_cells:
            bottom_row = row + cell.span[0]
            right_col = col + cell.span[1]
            self._set_rect_borders(row, bottom_row, col, right_col, cell.style)
            self._set_inner_borders(
                row, bottom_row, col, right_col, self.no_border
            )

    def _set_inner_borders(
        self,
        top_row: int,
        bottom_row: int,
        left_col: int,
        right_col: int,
        border_style: BorderStyle,
    ):
        """Set `border_style` to the inner borders of the rectangle (top_row,
        bottom_row, ...)
        """
        if bottom_row - top_row > 1:
            for col in range(left_col, right_col):
                for row in range(top_row + 1, bottom_row):
                    self.set_x_border(row, col, border_style)
        if right_col - left_col > 1:
            for row in range(top_row, bottom_row):
                for col in range(left_col + 1, right_col):
                    self.set_y_border(row, col, border_style)

    def _set_rect_borders(
        self,
        top_row: int,
        bottom_row: int,
        left_col: int,
        right_col: int,
        style: CellStyle,
    ):
        """Set border `style` to the rectangle (top_row, bottom_row, ...)

        The values describing the grid lines between the cells, see doc-strings
        for methods set_x_border() and set_y_border() and see comments for
        self._x_borders and self._y_borders.
        """
        for col in range(left_col, right_col):
            self.set_x_border(top_row, col, style.top)
            self.set_x_border(bottom_row, col, style.bottom)
        for row in range(top_row, bottom_row):
            self.set_y_border(row, left_col, style.left)
            self.set_y_border(row, right_col, style.right)

    def _set_frames(self, frames: Iterable[Frame]):
        """Set borders for all defined frames."""
        for frame in frames:
            top_row = frame.pos[0]
            left_col = frame.pos[1]
            bottom_row = top_row + frame.span[0]
            right_col = left_col + frame.span[1]
            self._set_rect_borders(
                top_row, bottom_row, left_col, right_col, frame.style
            )

    def _render_borders(self, layout: GenericLayoutType, table: TablePainter):
        """Render the grid lines as LINE entities into layout object."""

        def render_line(start: Vec2, end: Vec2, style: BorderStyle):
            """Render the LINE entity into layout object."""
            if style.status:
                layout.add_line(
                    start=start,
                    end=end,
                    dxfattribs={
                        "layer": layer,
                        "color": style.color,
                        "linetype": style.linetype,
                        "lineweight": style.lineweight,
                    },
                )

        def render_x_borders():
            """Draw the horizontal grid lines."""
            for row in range(table.nrows + 1):
                y = self.row_pos[row]
                for col in range(table.ncols):
                    left = self.col_pos[col]
                    right = self.col_pos[col + 1]
                    style = self.get_x_border(row, col)
                    render_line(Vec2(left, y), Vec2(right, y), style)

        def render_y_borders():
            """Draw the vertical grid lines."""
            for col in range(table.ncols + 1):
                x = self.col_pos[col]
                for row in range(table.nrows):
                    top = self.row_pos[row]
                    bottom = self.row_pos[row + 1]
                    style = self.get_y_border(row, col)
                    render_line(Vec2(x, top), Vec2(x, bottom), style)

        layer = table.grid_layer_name
        render_x_borders()
        render_y_borders()


class Frame:
    """Represent a rectangle cell area enclosed by borderlines.

    Args:
         table: the assigned data table
         pos: tuple (row, col), border goes left and top of pos
         span: count of cells that Frame covers, border goes right and below of this cells
         style: style name as string
    """

    def __init__(
        self,
        table: TablePainter,
        pos: tuple[int, int] = (0, 0),
        span: tuple[int, int] = (1, 1),
        style="default",
    ):
        self.table = table
        self.pos = pos
        self.span = span
        self.stylename = style

    @property
    def style(self) -> CellStyle:
        return self.table.get_cell_style(self.stylename)


class Cell:
    """Base class for table cells.

    Args:
        table: assigned data table
        style: style name as string
        span: tuple(spanrows, spancols), count of cells that cell covers

    A cell doesn't know its own position in the data table, because a cell can
    be used multiple times in the same or in different tables, therefore the
    cell itself can not determine if the cell-range reaches beyond the table
    borders.
    """

    def __init__(
        self,
        table: TablePainter,
        style="default",
        span: tuple[int, int] = (1, 1),
    ):
        self.table = table
        self.stylename = style
        # span values has to be >= 1
        self.span = span

    @property
    def span(self) -> tuple[int, int]:
        """Get/set table span parameters."""
        return self._span

    @span.setter
    def span(self, value: tuple[int, int]):
        """Ensures that span values are >= 1 in each direction."""
        self._span = (max(1, value[0]), max(1, value[1]))

    @property
    def style(self) -> CellStyle:
        """Returns the associated :class:`CellStyle`."""
        return self.table.get_cell_style(self.stylename)

    def render(
        self, layout: GenericLayoutType, coords: Sequence[float], layer: str
    ):
        """Renders the cell content into the given `layout`."""
        pass

    def get_workspace_coords(self, coords: Sequence[float]) -> Sequence[float]:
        """Reduces the cell-coords about the margin_x and the margin_y values."""
        margin_x = self.style.margin_x
        margin_y = self.style.margin_y
        return (
            coords[0] + margin_x,  # left
            coords[1] - margin_x,  # right
            coords[2] - margin_y,  # top
            coords[3] + margin_y,  # bottom
        )


CustomCell = Cell


class TextCell(Cell):
    """Implements a cell type containing a multi-line text. Uses the
    :class:`~ezdxf.addons.MTextSurrogate` add-on to render the multi-line
    text, therefore the content of these cells is compatible to DXF R12.

    Args:
        table: assigned data table
        text: multi line text, lines separated by the new line character ``"\\n"``
        style: cell style name as string
        span: tuple(rows, cols) area of cells to cover

    """

    def __init__(
        self,
        table: TablePainter,
        text: str,
        style="default",
        span: tuple[int, int] = (1, 1),
    ):
        super(TextCell, self).__init__(table, style, span)
        self.text = text

    def render(
        self, layout: GenericLayoutType, coords: Sequence[float], layer: str
    ):
        """Text cell.

        Args:
            layout: target layout
            coords: tuple of border-coordinates: left, right, top, bottom
            layer: target layer name as string

        """
        if not len(self.text):
            return

        left, right, top, bottom = self.get_workspace_coords(coords)
        style = self.style
        h_align, v_align = style.get_text_align_flags()
        rotated = self.style.rotation
        text = self.text
        if style.stacked:
            rotated = 0.0
            text = "\n".join((char for char in self.text.replace("\n", " ")))
        xpos = (left, float(left + right) / 2.0, right)[h_align]
        ypos = (bottom, float(bottom + top) / 2.0, top)[v_align - 1]
        mtext = MTextSurrogate(
            text,
            (xpos, ypos),
            line_spacing=self.style.line_spacing,
            style=self.style.text_style,
            char_height=self.style.char_height,
            rotation=rotated,
            width_factor=self.style.scale_x,
            align=style.align,
            color=self.style.text_color,
            layer=layer,
        )
        mtext.render(layout)


class BlockCell(Cell):
    """Implements a cell type containing a block reference.

    Args:
        table: table object
        blockdef: :class:`ezdxf.layouts.BlockLayout` instance
        attribs: BLOCK attributes as (tag, value) dictionary
        style: cell style name as string
        span: tuple(rows, cols) area of cells to cover

    """

    def __init__(
        self,
        table: TablePainter,
        blockdef: BlockLayout,
        style="default",
        attribs=None,
        span: tuple[int, int] = (1, 1),
    ):
        if attribs is None:
            attribs = {}
        super(BlockCell, self).__init__(table, style, span)
        self.block_name = blockdef.name  # dxf block name!
        self.attribs = attribs

    def render(
        self, layout: GenericLayoutType, coords: Sequence[float], layer: str
    ):
        """Create the cell content as INSERT-entity with trailing ATTRIB-Entities.

        Args:
            layout: target layout
            coords: tuple of border-coordinates : left, right, top, bottom
            layer: target layer name as string

        """
        left, right, top, bottom = self.get_workspace_coords(coords)
        style = self.style
        h_align, v_align = style.get_text_align_flags()
        xpos = (left, float(left + right) / 2.0, right)[h_align]
        ypos = (bottom, float(bottom + top) / 2.0, top)[v_align - 1]
        layout.add_auto_blockref(
            name=self.block_name,
            insert=(xpos, ypos),
            values=self.attribs,
            dxfattribs={
                "xscale": style.scale_x,
                "yscale": style.scale_y,
                "rotation": style.rotation,
                "layer": layer,
            },
        )


def sum_fields(
    start_value: float,
    fields: list[float],
    append: Callable[[float], None],
    sign: float = 1.0,
):
    """Adds step-by-step the fields-values, starting with <start_value>,
    and appends the resulting values to another object with the
    append-method.
    """
    position = start_value
    append(position)
    for element in fields:
        position += element * sign
        append(position)



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License

from ezdxf.version import __version__, version


def test_version():
    assert isinstance(__version__, str)
    assert isinstance(version, tuple)



# Copyright (c) 2019-2023 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.underlay import PdfUnderlay, DwfUnderlay, DgnUnderlay
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

PDFUNDERLAY = """0
PDFUNDERLAY
5
0
330
0
100
AcDbEntity
8
0
100
AcDbUnderlayReference
340
0
10
0.0
20
0.0
30
0.0
41
1.0
42
1.0
43
1.0
50
0.0
280
2
281
100
282
0
"""


@pytest.fixture
def entity():
    return PdfUnderlay.from_text(PDFUNDERLAY)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "PDFUNDERLAY" in ENTITY_CLASSES


def test_default_init():
    entity = PdfUnderlay()
    assert entity.dxftype() == "PDFUNDERLAY"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = PdfUnderlay.new(handle="ABBA", owner="0", dxfattribs={})
    assert entity.dxf.layer == "0"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.dxf.underlay_def_handle is None  # set by add_underlay()
    assert entity.dxf.scale_x == 1
    assert entity.dxf.scale_y == 1
    assert entity.dxf.scale_y == 1
    assert entity.dxf.rotation == 0
    assert entity.dxf.extrusion == (0, 0, 1)
    assert entity.dxf.flags == 10
    assert entity.dxf.contrast == 100
    assert entity.dxf.fade == 0
    assert entity.boundary_path == []


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.dxf.underlay_def_handle == "0"
    assert entity.dxf.scale_x == 1
    assert entity.dxf.scale_y == 1
    assert entity.dxf.scale_y == 1
    assert entity.dxf.rotation == 0
    assert entity.dxf.extrusion == (0, 0, 1)
    assert entity.dxf.flags == 2
    assert entity.dxf.contrast == 100
    assert entity.dxf.fade == 0
    assert entity.boundary_path == []


def test_write_dxf():
    entity = PdfUnderlay.from_text(PDFUNDERLAY)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(PDFUNDERLAY)
    assert result == expected


def test_dwf_underlay():
    assert DwfUnderlay.DXFTYPE == "DWFUNDERLAY"


def test_dgn_underlay():
    assert DgnUnderlay.DXFTYPE == "DGNUNDERLAY"



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Sequence,
    Union,
    Iterator,
    Optional,
)
from typing_extensions import Self
import array
import copy
from itertools import chain
from contextlib import contextmanager

from ezdxf.audit import AuditError
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER,
    DXF2000,
    DXFValueError,
    DXFStructureError,
    DXFIndexError,
)
from ezdxf.lldxf.packedtags import VertexArray, TagArray, TagList
from ezdxf.math import Matrix44, UVec, Vec3
from ezdxf.tools import take2
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace, DXFEntity
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.lldxf.tags import Tags
    from ezdxf.audit import Auditor

__all__ = ["Mesh", "MeshData"]

acdb_mesh = DefSubclass(
    "AcDbSubDMesh",
    {
        "version": DXFAttr(71, default=2),
        "blend_crease": DXFAttr(
            72,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # 0 is no smoothing
        "subdivision_levels": DXFAttr(
            91,
            default=0,
            validator=validator.is_greater_or_equal_zero,
            fixer=RETURN_DEFAULT,
        ),
        # 92: Vertex count of level 0
        # 10: Vertex position, multiple entries
        # 93: Size of face list of level 0
        # 90: Face list item, >=3 possible
        #     90: length of face list
        #     90: 1st vertex index
        #     90: 2nd vertex index ...
        # 94: Edge count of level 0
        #     90: Vertex index of 1st edge
        #     90: Vertex index of 2nd edge
        # 95: Edge crease count of level 0
        #     95 same as 94, or how is the 'edge create value' associated to edge index
        # 140: Edge crease value
        #
        # Overriding properties: how does this work?
        # 90: Count of sub-entity which property has been overridden
        # 91: Sub-entity marker
        # 92: Count of property was overridden
        # 90: Property type
        #     0 = Color
        #     1 = Material
        #     2 = Transparency
        #     3 = Material mapper
    },
)
acdb_mesh_group_codes = group_code_mapping(acdb_mesh)


class EdgeArray(TagArray):
    DTYPE = "L"

    def __len__(self) -> int:
        return len(self.values) // 2

    def __iter__(self) -> Iterator[tuple[int, int]]:
        for edge in take2(self.values):
            yield edge

    def set_data(self, edges: Iterable[tuple[int, int]]) -> None:
        self.values = array.array(self.DTYPE, chain.from_iterable(edges))

    def export_dxf(self, tagwriter: AbstractTagWriter):
        # count = count of edges not tags!
        tagwriter.write_tag2(94, len(self.values) // 2)
        for index in self.values:
            tagwriter.write_tag2(90, index)


class FaceList(TagList):
    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self) -> Iterable[array.array]:
        return iter(self.values)

    def export_dxf(self, tagwriter: AbstractTagWriter):
        # count = count of tags not faces!
        tagwriter.write_tag2(93, self.tag_count())
        for face in self.values:
            tagwriter.write_tag2(90, len(face))
            for index in face:
                tagwriter.write_tag2(90, index)

    def tag_count(self) -> int:
        return len(self.values) + sum(len(f) for f in self.values)

    def set_data(self, faces: Iterable[Sequence[int]]) -> None:
        _faces = []
        for face in faces:
            _faces.append(face_to_array(face))
        self.values = _faces


def face_to_array(face: Sequence[int]) -> array.array:
    max_index = max(face)
    if max_index < 256:
        dtype = "B"
    elif max_index < 65536:
        dtype = "I"
    else:
        dtype = "L"
    return array.array(dtype, face)


def create_vertex_array(tags: Tags, start_index: int) -> VertexArray:
    vertex_tags = tags.collect_consecutive_tags(codes=(10,), start=start_index)
    return VertexArray(data=[t.value for t in vertex_tags])


def create_face_list(tags: Tags, start_index: int) -> FaceList:
    faces = FaceList()
    faces_list = faces.values
    face: list[int] = []
    counter = 0
    for tag in tags.collect_consecutive_tags(codes=(90,), start=start_index):
        if not counter:
            # leading counter tag
            counter = tag.value
            if face:
                # group code 90 = 32 bit integer
                faces_list.append(face_to_array(face))
                face = []
        else:
            # followed by count face tags
            counter -= 1
            face.append(tag.value)

    # add last face
    if face:
        # group code 90 = 32 bit integer
        faces_list.append(face_to_array(face))

    return faces


def create_edge_array(tags: Tags, start_index: int) -> EdgeArray:
    return EdgeArray(data=collect_values(tags, start_index, code=90))  # int values


def collect_values(
    tags: Tags, start_index: int, code: int
) -> Iterable[Union[float, int]]:
    values = tags.collect_consecutive_tags(codes=(code,), start=start_index)
    return (t.value for t in values)


def create_crease_array(tags: Tags, start_index: int) -> array.array:
    return array.array("f", collect_values(tags, start_index, code=140))  # float values


COUNT_ERROR_MSG = "'MESH (#{}) without {} count.'"


@register_entity
class Mesh(DXFGraphic):
    """DXF MESH entity"""

    DXFTYPE = "MESH"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_mesh)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def __init__(self):
        super().__init__()
        self._vertices = VertexArray()  # vertices stored as array.array('d')
        self._faces = FaceList()  # face lists data
        self._edges = EdgeArray()  # edge indices stored as array.array('L')
        self._creases = array.array("f")  # creases stored as array.array('f')

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy data: vertices, faces, edges, creases."""
        assert isinstance(entity, Mesh)
        entity._vertices = copy.deepcopy(self._vertices)
        entity._faces = copy.deepcopy(self._faces)
        entity._edges = copy.deepcopy(self._edges)
        entity._creases = copy.deepcopy(self._creases)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(2)
            if tags:
                # Load mesh data and remove their tags from subclass
                self.load_mesh_data(tags, dxf.handle)
                # Load remaining data into name space
                processor.fast_load_dxfattribs(
                    dxf, acdb_mesh_group_codes, 2, recover=True
                )
            else:
                raise DXFStructureError(
                    f"missing 'AcDbSubMesh' subclass in MESH(#{dxf.handle})"
                )
        return dxf

    def load_mesh_data(self, mesh_tags: Tags, handle: str) -> None:
        def process_vertices():
            try:
                vertex_count_index = mesh_tags.tag_index(92)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, "vertex"))
            vertices = create_vertex_array(mesh_tags, vertex_count_index + 1)
            # Remove vertex count tag and all vertex tags
            end_index = vertex_count_index + 1 + len(vertices)
            del mesh_tags[vertex_count_index:end_index]
            return vertices

        def process_faces():
            try:
                face_count_index = mesh_tags.tag_index(93)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, "face"))
            else:
                # Remove face count tag and all face tags
                faces = create_face_list(mesh_tags, face_count_index + 1)
                end_index = face_count_index + 1 + faces.tag_count()
                del mesh_tags[face_count_index:end_index]
                return faces

        def process_edges():
            try:
                edge_count_index = mesh_tags.tag_index(94)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, "edge"))
            else:
                edges = create_edge_array(mesh_tags, edge_count_index + 1)
                # Remove edge count tag and all edge tags
                end_index = edge_count_index + 1 + len(edges.values)
                del mesh_tags[edge_count_index:end_index]
                return edges

        def process_creases():
            try:
                crease_count_index = mesh_tags.tag_index(95)
            except DXFValueError:
                raise DXFStructureError(COUNT_ERROR_MSG.format(handle, "crease"))
            else:
                creases = create_crease_array(mesh_tags, crease_count_index + 1)
                # Remove crease count tag and all crease tags
                end_index = crease_count_index + 1 + len(creases)
                del mesh_tags[crease_count_index:end_index]
                return creases

        self._vertices = process_vertices()
        self._faces = process_faces()
        self._edges = process_edges()
        self._creases = process_creases()

    def preprocess_export(self, tagwriter: AbstractTagWriter) -> bool:
        """Pre requirement check and pre-processing for export.

        Returns ``False``  if entity should not be exported at all.

            - MESH without vertices creates an invalid DXF file for AutoCAD and BricsCAD.
            - MESH without faces creates an invalid DXF file for AutoCAD and BricsCAD.

        (internal API)
        """
        if len(self.vertices) == 0 or len(self.faces) == 0:
            return False
        return True

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_mesh.name)
        self.dxf.export_dxf_attribs(
            tagwriter, ["version", "blend_crease", "subdivision_levels"]
        )
        self.export_mesh_data(tagwriter)
        self.export_override_data(tagwriter)

    def export_mesh_data(self, tagwriter: AbstractTagWriter):
        tagwriter.write_tag2(92, len(self.vertices))
        self._vertices.export_dxf(tagwriter, code=10)
        self._faces.export_dxf(tagwriter)
        self._edges.export_dxf(tagwriter)

        creases = self._fixed_crease_values()
        tagwriter.write_tag2(95, len(self.creases))
        for crease_value in creases:
            tagwriter.write_tag2(140, crease_value)

    def _fixed_crease_values(self) -> list[float]:
        # The edge count has to match the crease count, otherwise its an invalid
        # DXF file to AutoCAD!
        edge_count = len(self._edges)
        creases = list(self.creases)
        crease_count = len(creases)
        if edge_count < crease_count:
            creases = creases[:edge_count]
        while edge_count > len(creases):
            creases.append(0.0)
        return creases

    def export_override_data(self, tagwriter: AbstractTagWriter):
        tagwriter.write_tag2(90, 0)

    @property
    def creases(self) -> array.array:
        """Creases as :class:`array.array`. (read/write)"""
        return self._creases

    @creases.setter
    def creases(self, values: Iterable[float]) -> None:
        self._creases = array.array("f", values)

    @property
    def vertices(self):
        """Vertices as list like :class:`~ezdxf.lldxf.packedtags.VertexArray`.
        (read/write)
        """
        return self._vertices

    @vertices.setter
    def vertices(self, points: Iterable[UVec]) -> None:
        self._vertices = VertexArray(points)

    @property
    def edges(self):
        """Edges as list like :class:`~ezdxf.lldxf.packedtags.TagArray`.
        (read/write)
        """
        return self._edges

    @edges.setter
    def edges(self, edges: Iterable[tuple[int, int]]) -> None:
        self._edges.set_data(edges)

    @property
    def faces(self):
        """Faces as list like :class:`~ezdxf.lldxf.packedtags.TagList`.
        (read/write)
        """
        return self._faces

    @faces.setter
    def faces(self, faces: Iterable[Sequence[int]]) -> None:
        self._faces.set_data(faces)

    def get_data(self) -> MeshData:
        return MeshData(self)

    def set_data(self, data: MeshData) -> None:
        self.vertices = data.vertices
        self._faces.set_data(data.faces)
        self._edges.set_data(data.edges)
        self.creases = array.array("f", data.edge_crease_values)
        if len(self.edges) != len(self.creases):
            raise DXFValueError("count of edges must match count of creases")

    @contextmanager
    def edit_data(self) -> Iterator[MeshData]:
        """Context manager for various mesh data, returns a :class:`MeshData` instance.

        Despite that vertices, edge and faces are accessible as packed data types, the
        usage of :class:`MeshData` by context manager :meth:`edit_data` is still
        recommended.
        """
        data = self.get_data()
        yield data
        self.set_data(data)

    def transform(self, m: Matrix44) -> Mesh:
        """Transform the MESH entity by transformation matrix `m` inplace."""
        self._vertices.transform(m)
        self.post_transform(m)
        return self

    def audit(self, auditor: Auditor) -> None:
        if not self.is_alive:
            return
        super().audit(auditor)
        if len(self.edges) != len(self.creases):
            self.creases = self._fixed_crease_values()  # type: ignore
            auditor.fixed_error(
                code=AuditError.INVALID_CREASE_VALUE_COUNT,
                message=f"fixed invalid count of crease values in {str(self)}",
                dxf_entity=self,
            )
        if len(self.vertices) == 0 or len(self.faces) == 0:
            # A MESH without vertices or vertices creates an invalid DXF file for
            # AutoCAD and BricsCAD
            auditor.fixed_error(
                code=AuditError.INVALID_VERTEX_COUNT,
                message=f"removed {str(self)} without vertices or faces",
            )
            auditor.trash(self)


class MeshData:
    def __init__(self, mesh: Mesh) -> None:
        self.vertices: list[Vec3] = Vec3.list(mesh.vertices)
        self.faces: list[Sequence[int]] = list(mesh.faces)
        self.edges: list[tuple[int, int]] = list(mesh.edges)
        self.edge_crease_values: list[float] = list(mesh.creases)

    def add_face(self, vertices: Iterable[UVec]) -> Sequence[int]:
        """Add a face by a list of vertices."""
        indices = tuple(self.add_vertex(vertex) for vertex in vertices)
        self.faces.append(indices)
        return indices

    def add_edge_crease(self, v1: int, v2: int, crease: float):
        """Add an edge crease value, the edge is defined by the vertex indices
        `v1` and `v2`.

        The crease value defines the amount of subdivision that will be applied
        to this edge.  A crease value of the subdivision level prevents the edge from
        deformation and a value of 0.0 means no protection from subdividing.

        """
        if v1 < 0 or v1 > len(self.vertices):
            raise DXFIndexError("vertex index `v1` out of range")
        if v2 < 0 or v2 > len(self.vertices):
            raise DXFIndexError("vertex index `v2` out of range")
        self.edges.append((v1, v2))
        self.edge_crease_values.append(crease)

    def add_vertex(self, vertex: UVec) -> int:
        if len(vertex) != 3:
            raise DXFValueError("Parameter vertex has to be a 3-tuple (x, y, z).")
        index = len(self.vertices)
        self.vertices.append(Vec3(vertex))
        return index

    def optimize(self):
        """Reduce vertex count by merging coincident vertices."""

        def merge_coincident_vertices() -> dict[int, int]:
            original_vertices = [
                (v.xyz, index, v) for index, v in enumerate(self.vertices)
            ]
            original_vertices.sort()
            self.vertices = []
            index_map: dict[int, int] = {}
            prev_vertex = sentinel = Vec3()
            index = 0
            for _, original_index, vertex in original_vertices:
                if prev_vertex is sentinel or not vertex.isclose(prev_vertex):
                    index = len(self.vertices)
                    self.vertices.append(vertex)
                    index_map[original_index] = index
                    prev_vertex = vertex
                else:  # this is a coincident vertex
                    index_map[original_index] = index
            return index_map

        def remap_faces() -> None:
            self.faces = remap_indices(self.faces)

        def remap_edges() -> None:
            self.edges = remap_indices(self.edges)  # type: ignore

        def remap_indices(indices: Sequence[Sequence[int]]) -> list[Sequence[int]]:
            mapped_indices: list[Sequence[int]] = []
            for entry in indices:
                mapped_indices.append(tuple(index_map[index] for index in entry))
            return mapped_indices

        index_map = merge_coincident_vertices()
        remap_faces()
        remap_edges()



# Copyright (c) 2018-2022 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    Dict,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing_extensions import TypeAlias
from ezdxf.math import (
    BoundingBox,
    Matrix44,
    NULLVEC,
    OCS,
    UCS,
    UVec,
    Vec3,
    area,
    is_planar_face,
    normal_vector_3p,
    safe_normal_vector,
    subdivide_face,
    subdivide_ngons,
)

if TYPE_CHECKING:
    from ezdxf.entities import Polyface, Polymesh, Mesh, Solid3d
    from ezdxf.eztypes import GenericLayoutType

T = TypeVar("T")


class EdgeStat(NamedTuple):
    """Named tuple of edge statistics."""

    count: int  # type: ignore
    balance: int


Face: TypeAlias = Sequence[int]
Edge: TypeAlias = Tuple[int, int]
EdgeStats: TypeAlias = Dict[Edge, EdgeStat]


class MeshBuilderError(Exception):
    pass


class NodeMergingError(MeshBuilderError):
    pass


class MultipleMeshesError(MeshBuilderError):
    pass


class DegeneratedPathError(MeshBuilderError):
    pass


class NonManifoldMeshError(MeshBuilderError):
    pass


def open_faces(faces: Iterable[Face]) -> Iterator[Face]:
    """Yields all faces with more than two vertices as open faces
    (first vertex index != last vertex index).
    """
    for face in faces:
        if len(face) < 3:
            continue
        if face[0] == face[-1]:
            yield face[:-1]
        else:
            yield face


def normalize_faces(
    faces: list[Sequence[int]],
    *,
    close=False,
) -> Iterator[Sequence[int]]:
    """Removes duplicated vertices and returns closed or open faces according
    the `close` argument. Returns only faces with at least 3 edges.
    """
    for face in open_faces(faces):
        new_face = [face[0]]
        for index in face[1:]:
            if new_face[-1] != index:
                new_face.append(index)
        if len(new_face) < 3:
            continue
        if close:
            new_face.append(new_face[0])
        yield tuple(new_face)


def all_edges(faces: Iterable[Face]) -> Iterator[Edge]:
    """Yields all face edges as int tuples."""
    for face in open_faces(faces):
        yield from face_edges(face)


def face_edges(face: Face) -> Iterable[Edge]:
    """Yields all edges of a single open face as int tuples."""
    size = len(face)
    for index in range(size):
        yield face[index], face[(index + 1) % size]


def get_edge_stats(faces: Iterable[Face]) -> EdgeStats:
    """Returns the edge statistics.

    The Edge statistic contains for each edge `(a, b)` the :class:`EdgeStat` as
    tuple `(count, balance)` where the vertex index `a` is always smaller than
    the vertex index `b`.

    The edge count is how often this edge is used in faces as `(a, b)` or
    `(b, a)` and the balance is the count of edge `(a, b)` minus the count of
    edge `(b, a)` and should be 0 in "healthy" closed surfaces.
    A balance not 0 indicates an error which may be double coincident faces or
    mixed face vertex orders.

    """
    new_edge = EdgeStat(0, 0)
    stats: EdgeStats = {}
    for a, b in all_edges(faces):
        edge = a, b
        orientation = +1
        if a > b:
            edge = b, a
            orientation = -1
        # for all edges: count should be 2 and balance should be 0
        count, balance = stats.get(edge, new_edge)
        stats[edge] = EdgeStat(count + 1, balance + orientation)
    return stats


def estimate_face_normals_direction(
    vertices: Sequence[Vec3], faces: Sequence[Face]
) -> float:
    """Returns the estimated face-normals direction as ``float`` value
    in the range [-1.0, 1.0] for a closed surface.

    This heuristic works well for simple convex hulls but struggles with
    more complex structures like a torus (doughnut).

    A counter-clockwise (ccw) vertex arrangement is assumed but a
    clockwise (cw) arrangement works too but the values are reversed.

    The closer the value to 1.0 (-1.0 for cw) the more likely all normals
    pointing outwards from the surface.

    The closer the value to -1.0 (1.0 for cw) the more likely all normals
    pointing inwards from the surface.

    """
    n_vertices = len(vertices)
    if n_vertices == 0:
        return 0.0

    mesh_centroid = Vec3.sum(vertices) / n_vertices
    count = 0
    direction_sum = 0.0
    for face in faces:
        if len(face) < 3:
            continue
        try:
            face_vertices = tuple(vertices[i] for i in face)
        except IndexError:
            continue
        face_centroid = Vec3.sum(face_vertices) / len(face)
        try:
            face_normal = normal_vector_3p(
                face_vertices[0], face_vertices[1], face_vertices[2]
            )
        except ZeroDivisionError:
            continue
        try:
            outward_vec = (face_centroid - mesh_centroid).normalize()
        except ZeroDivisionError:
            continue
        direction_sum += face_normal.dot(outward_vec)
        count += 1
    if count > 0:
        return direction_sum / count
    return 0.0


def flip_face_normals(
    faces: Sequence[Sequence[int]],
) -> Iterator[Sequence[int]]:
    for face in faces:
        yield tuple(reversed(face))


def volume6(a: Vec3, b: Vec3, c: Vec3) -> float:
    """
    Returns six times the volume of the tetrahedron determined by abc
    and the origin (0, 0, 0).
    The volume is positive if the origin is on the negative side of abc,
    where the positive side is determined by the right-hand--rule.
    So the volume is positive if the ccw normal to abc points outside the
    tetrahedron.

    This code is taken from chull.c; see "Computational Geometry in C."
    """
    return (
        a.z * (b.x * c.y - b.y * c.x)
        + a.y * (b.z * c.x - b.x * c.z)
        + a.x * (b.y * c.z - b.z * c.y)
    )


def area_3d(polygon: Sequence[Vec3]) -> float:
    try:
        ocs = OCS(safe_normal_vector(polygon))
    except ZeroDivisionError:
        return 0.0
    return area(ocs.points_from_wcs(polygon))


# Mesh Topology Analysis using the Euler Characteristic
# https://max-limper.de/publications/Euler/index.html


class MeshDiagnose:
    def __init__(self, mesh: MeshBuilder):
        self._mesh = mesh
        self._edge_stats: EdgeStats = {}
        self._bbox = BoundingBox()
        self._face_normals: list[Vec3] = []

    @property
    def vertices(self) -> Sequence[Vec3]:
        """Sequence of mesh vertices as :class:`~ezdxf.math.Vec3` instances"""
        return self._mesh.vertices

    @property
    def faces(self) -> Sequence[Face]:
        """Sequence of faces as ``Sequence[int]``"""
        return self._mesh.faces

    @property
    def face_normals(self) -> Sequence[Vec3]:
        """Returns all face normal vectors  as sequence. The ``NULLVEC``
        instance is used as normal vector for degenerated faces. (cached data)

        """
        if len(self._face_normals) == 0:
            self._face_normals = list(self._mesh.face_normals())
        return self._face_normals

    @property
    def bbox(self) -> BoundingBox:
        """Returns the :class:`~ezdxf.math.BoundingBox` of the mesh. (cached data)"""
        if not self._bbox.has_data:
            self._bbox = self._mesh.bbox()
        return self._bbox

    @property
    def n_vertices(self) -> int:
        """Returns the vertex count."""
        return len(self.vertices)

    @property
    def n_faces(self) -> int:
        """Returns the face count."""
        return len(self.faces)

    @property
    def n_edges(self) -> int:
        """Returns the unique edge count. (cached data)"""
        return len(self.edge_stats)

    @property
    def edge_stats(self) -> EdgeStats:
        """Returns the edge statistics as a ``dict``. The dict-key is the edge
        as tuple of two vertex indices `(a, b)` where `a` is always smaller than
        `b`. The dict-value is an :class:`EdgeStat` tuple of edge count and edge
        balance, see :class:`EdgeStat` for the definition of edge count and
        edge balance. (cached data)
        """
        if len(self._edge_stats) == 0:
            self._edge_stats = get_edge_stats(self.faces)
        return self._edge_stats

    @property
    def euler_characteristic(self) -> int:
        """Returns the Euler characteristic:
        https://en.wikipedia.org/wiki/Euler_characteristic

        This number is always 2 for convex polyhedra.
        """
        return self.n_vertices - self.n_edges + self.n_faces

    @property
    def is_edge_balance_broken(self) -> bool:
        """Returns ``True`` if the edge balance is broken, this indicates a
        topology error for closed surfaces. A non-broken edge balance reflects
        that each edge connects two faces, where the edge is clockwise oriented
        in the first face and counter-clockwise oriented in the second face.
        A broken edge balance indicates possible topology errors like mixed
        face vertex orientations or a non-manifold mesh where an edge connects
        more than two faces. (cached data)

        """
        return any(e.balance != 0 for e in self.edge_stats.values())

    @property
    def is_manifold(self) -> bool:
        """Returns ``True`` if all edges have an edge count < 3. (cached data)

        A non-manifold mesh has edges with 3 or more connected faces.

        """
        return all(edge.count < 3 for edge in self.edge_stats.values())

    @property
    def is_closed_surface(self) -> bool:
        """Returns ``True`` if the mesh has a closed surface.
        This method does not require a unified face orientation.
        If multiple separated meshes are present the state is only ``True`` if
        **all** meshes have a closed surface. (cached data)

        Returns ``False`` for non-manifold meshes.

        """
        return all(edge.count == 2 for edge in self.edge_stats.values())

    def total_edge_count(self) -> int:
        """Returns the total edge count of all faces, shared edges are counted
        separately for each face. In closed surfaces this count should be 2x
        the unique edge count :attr:`n_edges`. (cached data)
        """
        return sum(e.count for e in self.edge_stats.values())

    def unique_edges(self) -> Iterable[Edge]:
        """Yields the unique edges of the mesh as int 2-tuples. (cached data)"""
        return self.edge_stats.keys()

    def estimate_face_normals_direction(self) -> float:
        """Returns the estimated face-normals direction as ``float`` value
        in the range [-1.0, 1.0] for a closed surface.

        This heuristic works well for simple convex hulls but struggles with
        more complex structures like a torus (doughnut).

        A counter-clockwise (ccw) vertex arrangement for outward pointing faces
        is assumed but a clockwise (cw) arrangement works too but the return
        values are reversed.

        The closer the value to 1.0 (-1.0 for cw) the more likely all normals
        pointing outwards from the surface.

        The closer the value to -1.0 (1.0 for cw) the more likely all normals
        pointing inwards from the surface.

        There are no exact confidence values if all faces pointing outwards,
        here some examples for surfaces created by :mod:`ezdxf.render.forms`
        functions:

            - :func:`~ezdxf.render.forms.cube` returns 1.0
            - :func:`~ezdxf.render.forms.cylinder` returns 0.9992
            - :func:`~ezdxf.render.forms.sphere` returns 0.9994
            - :func:`~ezdxf.render.forms.cone` returns 0.9162
            - :func:`~ezdxf.render.forms.cylinder` with all hull faces pointing
              outwards but caps pointing inwards returns 0.7785 but the
              property :attr:`is_edge_balance_broken` returns ``True`` which
              indicates the mixed vertex orientation
            - and the estimation of 0.0469 for a :func:`~ezdxf.render.forms.torus`
              is barely usable

        """
        return estimate_face_normals_direction(self.vertices, self.faces)

    def has_non_planar_faces(self) -> bool:
        """Returns ``True`` if any face is non-planar."""
        return not all(is_planar_face(face) for face in self._mesh.faces_as_vertices())

    def volume(self) -> float:
        """Returns the volume of a closed surface or 0 otherwise.

        .. warning::

            The face vertices have to be in counter-clockwise order, this
            requirement is not checked by this method.

            The result is not correct for multiple separated meshes in a single
            MeshBuilder object!!!

        """
        if self.is_closed_surface:
            volume = 0.0
            for face in self._mesh.tessellation(3):
                volume += volume6(face[0], face[1], face[2])
            return volume / 6.0
        return 0.0

    def surface_area(self) -> float:
        """Returns the surface area."""
        v = self.vertices
        surface_area = 0.0
        for face in self._mesh.open_faces():
            polygon = [v[i] for i in face]
            surface_area += area_3d(polygon)
        return surface_area

    def centroid(self) -> Vec3:
        """Returns the centroid of all vertices. (center of mass)"""
        if self.n_vertices > 0:
            return Vec3.sum(self.vertices) / self.n_vertices
        return NULLVEC


class MeshBuilder:
    """A simple Mesh builder. Stores a list of vertices and a faces list where
    each face is a list of indices into the vertices list.

    The :meth:`render_mesh` method, renders the mesh into a DXF MESH entity.
    The MESH entity supports ngons in AutoCAD, ngons are polygons with more
    than 4 vertices.

    Can only create new meshes.

    """

    def __init__(self) -> None:
        self.vertices: list[Vec3] = []
        # face storage, each face is a tuple of vertex indices (v0, v1, v2, v3, ....)
        self.faces: list[Sequence[int]] = []

    def bbox(self) -> BoundingBox:
        """Returns the :class:`~ezdxf.math.BoundingBox` of the mesh."""
        return BoundingBox(self.vertices)

    def copy(self):
        """Returns a copy of mesh."""
        return self.from_builder(self)

    def diagnose(self) -> MeshDiagnose:
        """Returns the :class:`MeshDiagnose` object for this mesh."""
        return MeshDiagnose(self)

    def get_face_vertices(self, index: int) -> Sequence[Vec3]:
        """Returns the face `index` as sequence of :class:`~ezdxf.math.Vec3`
        objects.
        """
        vertices = self.vertices
        return tuple(vertices[vi] for vi in self.faces[index])

    def get_face_normal(self, index: int) -> Vec3:
        """Returns the normal vector of the face `index` as :class:`~ezdxf.math.Vec3`,
        returns the ``NULLVEC`` instance for degenerated  faces.
        """
        face = self.get_face_vertices(index)
        try:
            return safe_normal_vector(face)
        except (ValueError, ZeroDivisionError):
            return NULLVEC

    def face_normals(self) -> Iterator[Vec3]:
        """Yields all face normals, yields the ``NULLVEC`` instance for degenerated
        faces.
        """
        for face in self.faces_as_vertices():
            try:
                yield safe_normal_vector(face)
            except (ValueError, ZeroDivisionError):
                yield NULLVEC

    def faces_as_vertices(self) -> Iterator[list[Vec3]]:
        """Yields all faces as list of vertices."""
        v = self.vertices
        for face in self.faces:
            yield [v[index] for index in face]

    def open_faces(self) -> Iterator[Face]:
        """Yields all faces as sequence of integers where the first vertex
        is not coincident with the last vertex.
        """
        yield from open_faces(self.faces)

    def add_face(self, vertices: Iterable[UVec]) -> None:
        """Add a face as vertices list to the mesh. A face requires at least 3
        vertices, each vertex is a ``(x, y, z)`` tuple or
        :class:`~ezdxf.math.Vec3` object. The new vertex indices are stored as
        face in the :attr:`faces` list.

        Args:
            vertices: list of at least 3 vertices ``[(x1, y1, z1), (x2, y2, z2),
                (x3, y3, y3), ...]``

        """
        self.faces.append(self.add_vertices(vertices))

    def add_vertices(self, vertices: Iterable[UVec]) -> Face:
        """Add new vertices to the mesh, each vertex is a ``(x, y, z)`` tuple
        or a :class:`~ezdxf.math.Vec3` object, returns the indices of the
        `vertices` added to the :attr:`vertices` list.

        e.g. adding 4 vertices to an empty mesh, returns the indices
        ``(0, 1, 2, 3)``, adding additional 4 vertices returns the indices
        ``(4, 5, 6, 7)``.

        Args:
            vertices: list of vertices, vertex as ``(x, y, z)`` tuple or
                :class:`~ezdxf.math.Vec3` objects

        Returns:
            tuple: indices of the `vertices` added to the :attr:`vertices` list

        """
        start_index = len(self.vertices)
        self.vertices.extend(Vec3.generate(vertices))
        return tuple(range(start_index, len(self.vertices)))

    def add_mesh(
        self,
        vertices: Optional[list[Vec3]] = None,
        faces: Optional[list[Face]] = None,
        mesh=None,
    ) -> None:
        """Add another mesh to this mesh.

        A `mesh` can be a :class:`MeshBuilder`, :class:`MeshVertexMerger` or
        :class:`~ezdxf.entities.Mesh` object or requires the attributes
        :attr:`vertices` and :attr:`faces`.

        Args:
            vertices: list of vertices, a vertex is a ``(x, y, z)`` tuple or
                :class:`~ezdxf.math.Vec3` object
            faces: list of faces, a face is a list of vertex indices
            mesh: another mesh entity

        """
        if mesh is not None:
            vertices = Vec3.list(mesh.vertices)
            faces = mesh.faces

        if vertices is None:
            raise ValueError("Requires vertices or another mesh.")
        faces = faces or []
        indices = self.add_vertices(vertices)

        for face_vertices in open_faces(faces):
            self.faces.append(tuple(indices[vi] for vi in face_vertices))

    def render_mesh(
        self,
        layout: GenericLayoutType,
        dxfattribs=None,
        matrix: Optional[Matrix44] = None,
        ucs: Optional[UCS] = None,
    ):
        """Render mesh as :class:`~ezdxf.entities.Mesh` entity into `layout`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            dxfattribs: dict of DXF attributes e.g. ``{'layer': 'mesh', 'color': 7}``
            matrix: transformation matrix of type :class:`~ezdxf.math.Matrix44`
            ucs: transform vertices by :class:`~ezdxf.math.UCS` to :ref:`WCS`

        """
        dxfattribs = dict(dxfattribs) if dxfattribs else {}
        vertices = self.vertices
        if matrix is not None:
            vertices = matrix.transform_vertices(vertices)
        if ucs is not None:
            vertices = ucs.points_to_wcs(vertices)  # type: ignore
        mesh = layout.add_mesh(dxfattribs=dxfattribs)
        with mesh.edit_data() as data:
            # data will be copied at setting in edit_data()
            # ignore edges and creases!
            data.vertices = list(vertices)
            data.faces = list(self.faces)
        return mesh

    def render_normals(
        self,
        layout: GenericLayoutType,
        length: float = 1,
        relative=True,
        dxfattribs=None,
    ):
        """Render face normals as :class:`~ezdxf.entities.Line` entities into
        `layout`, useful to check orientation of mesh faces.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            length: visual length of normal, use length < 0 to point normals in
                opposite direction
            relative: scale length relative to face size if ``True``
            dxfattribs: dict of DXF attributes e.g. ``{'layer': 'normals', 'color': 6}``

        """
        dxfattribs = dict(dxfattribs) if dxfattribs else {}
        for face in self.faces_as_vertices():
            count = len(face)
            if count < 3:
                continue
            center = Vec3.sum(face) / count
            try:
                n = safe_normal_vector(face)
            except ZeroDivisionError:
                continue
            if relative:
                _length = (face[0] - center).magnitude * length
            else:
                _length = length
            layout.add_line(center, center + n * _length, dxfattribs=dxfattribs)

    @classmethod
    def from_mesh(cls: Type[T], other: Union[MeshBuilder, Mesh]) -> T:
        """Create new mesh from other mesh as class method.

        Args:
            other: `mesh` of type :class:`MeshBuilder` and inherited or DXF
                :class:`~ezdxf.entities.Mesh` entity or any object providing
                attributes :attr:`vertices`, :attr:`edges` and :attr:`faces`.

        """
        # just copy properties
        mesh = cls()
        assert isinstance(mesh, MeshBuilder)
        mesh.add_mesh(mesh=other)
        return mesh  # type: ignore

    @classmethod
    def from_polyface(cls: Type[T], other: Union[Polymesh, Polyface]) -> T:
        """Create new mesh from a  :class:`~ezdxf.entities.Polyface` or
        :class:`~ezdxf.entities.Polymesh` object.

        """
        if other.dxftype() != "POLYLINE":
            raise TypeError(f"Unsupported DXF type: {other.dxftype()}")

        mesh = cls()
        assert isinstance(mesh, MeshBuilder)
        if other.is_poly_face_mesh:
            _, faces = other.indexed_faces()  # type: ignore
            for face in faces:
                mesh.add_face(face.points())
        elif other.is_polygon_mesh:
            vertices = other.get_mesh_vertex_cache()  # type: ignore
            for m in range(other.dxf.m_count - 1):
                for n in range(other.dxf.n_count - 1):
                    mesh.add_face(
                        (
                            vertices[m, n],
                            vertices[m, n + 1],
                            vertices[m + 1, n + 1],
                            vertices[m + 1, n],
                        )
                    )
        else:
            raise TypeError("Not a polymesh or polyface.")
        return mesh  # type: ignore

    def render_polyface(
        self,
        layout: GenericLayoutType,
        dxfattribs=None,
        matrix: Optional[Matrix44] = None,
        ucs: Optional[UCS] = None,
    ):
        """Render mesh as :class:`~ezdxf.entities.Polyface` entity into
        `layout`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            dxfattribs: dict of DXF attributes e.g. ``{'layer': 'mesh', 'color': 7}``
            matrix: transformation matrix of type :class:`~ezdxf.math.Matrix44`
            ucs: transform vertices by :class:`~ezdxf.math.UCS` to :ref:`WCS`

        """
        dxfattribs = dict(dxfattribs) if dxfattribs else {}
        polyface = layout.add_polyface(dxfattribs=dxfattribs)
        t = MeshTransformer.from_builder(self)
        if matrix is not None:
            t.transform(matrix)
        if ucs is not None:
            t.transform(ucs.matrix)
        polyface.append_faces(
            t.tessellation(max_vertex_count=4),
            dxfattribs=dxfattribs,
        )
        return polyface

    def render_3dsolid(self, layout: GenericLayoutType, dxfattribs=None) -> Solid3d:
        """Render mesh as :class:`~ezdxf.entities.Solid3d` entity into `layout`.

        This is an **experimental** feature to create simple 3DSOLID entities from 
        polyhedrons.
        
        The method supports closed and open shells.  A 3DSOLID entity can contain 
        multiple shells.  Separate the meshes beforehand by the method 
        :meth:`separate_meshes` if required.  The normals vectors of all faces should 
        point outwards. Faces can have more than 3 vertices (ngons) but non-planar 
        faces and concave faces will cause problems in some CAD applications.  The 
        method :meth:`mesh_tesselation` can help to break down the faces into triangles.

        Requires a valid DXF document for `layout` and DXF version R2000 or newer.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            dxfattribs: dict of DXF attributes e.g. ``{'layer': 'mesh', 'color': 7}``
        
        Raises:
            DXFValueError: valid DXF document required, if :attr:`layout.doc` is ``None``
            DXFVersionError: invalid DXF version

        .. versionadded:: 1.2.0

        """
        from ezdxf.acis import api as acis

        dxfattribs = dict(dxfattribs) if dxfattribs else {}
        solid3d = layout.add_3dsolid(dxfattribs=dxfattribs)
        body = acis.body_from_mesh(self)
        acis.export_dxf(solid3d, [body])
        return solid3d

    def render_3dfaces(
        self,
        layout: GenericLayoutType,
        dxfattribs=None,
        matrix: Optional[Matrix44] = None,
        ucs: Optional[UCS] = None,
    ):
        """Render mesh as :class:`~ezdxf.entities.Face3d` entities into
        `layout`.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            dxfattribs: dict of DXF attributes e.g. ``{'layer': 'mesh', 'color': 7}``
            matrix: transformation matrix of type :class:`~ezdxf.math.Matrix44`
            ucs: transform vertices by :class:`~ezdxf.math.UCS` to :ref:`WCS`

        """
        dxfattribs = dict(dxfattribs) if dxfattribs else {}
        t = MeshTransformer.from_builder(self)
        if matrix is not None:
            t.transform(matrix)
        if ucs is not None:
            t.transform(ucs.matrix)
        for face in t.tessellation(max_vertex_count=4):
            layout.add_3dface(face, dxfattribs=dxfattribs)

    @classmethod
    def from_builder(cls: Type[T], other: MeshBuilder) -> T:
        """Create new mesh from other mesh builder, faster than
        :meth:`from_mesh` but supports only :class:`MeshBuilder` and inherited
        classes.

        """
        # just copy properties
        mesh = cls()
        assert isinstance(mesh, MeshBuilder)

        # DO NOT COPY CACHES!
        mesh.vertices = list(other.vertices)
        mesh.faces = list(other.faces)
        return mesh  # type: ignore

    def merge_coplanar_faces(self, passes: int = 1) -> MeshTransformer:
        """Returns a new :class:`MeshBuilder` object with merged adjacent
        coplanar faces.

        The faces have to share at least two vertices and have to have the
        same clockwise or counter-clockwise vertex order.

        The current implementation is not very capable!

        """
        mesh = self
        for _ in range(passes):
            mesh = _merge_adjacent_coplanar_faces(mesh.vertices, mesh.faces)
        return MeshTransformer.from_builder(mesh)

    def subdivide(self, level: int = 1, quads=True) -> MeshTransformer:
        """Returns a new :class:`MeshTransformer` object with all faces subdivided.

        Args:
             level: subdivide levels from 1 to max of 5
             quads: create quad faces if ``True`` else create triangles
        """
        mesh = self
        level = min(int(level), 5)
        while level > 0:
            mesh = _subdivide(mesh, quads)
            level -= 1
        return MeshTransformer.from_builder(mesh)

    def optimize_vertices(self, precision: int = 6) -> MeshTransformer:
        """Returns a new mesh with optimized vertices. Coincident vertices are
        merged together and all faces are open faces (first vertex != last
        vertex). Uses internally the :class:`MeshVertexMerger` class to merge
        vertices.
        """
        mesh = MeshVertexMerger(precision=precision)
        mesh.add_mesh(mesh=self)
        return MeshTransformer.from_builder(mesh)

    def subdivide_ngons(self, max_vertex_count=4) -> Iterator[Sequence[Vec3]]:
        """Yields all faces as sequence of :class:`~ezdxf.math.Vec3` instances,
        where all ngons which have more than `max_vertex_count` vertices gets
        subdivided.
        In contrast to the :meth:`tessellation` method, creates this method a
        new vertex in the centroid of the face. This can create a more regular
        tessellation but only works reliable for convex faces!
        """
        yield from subdivide_ngons(self.faces_as_vertices(), max_vertex_count)

    def tessellation(self, max_vertex_count: int = 4) -> Iterator[Sequence[Vec3]]:
        """Yields all faces as sequence of :class:`~ezdxf.math.Vec3` instances,
        each face has no more vertices than the given `max_vertex_count`. This
        method uses the "ear clipping" algorithm which works with concave faces
        too and does not create any additional vertices.
        """
        from ezdxf.math.triangulation import mapbox_earcut_3d

        for face in self.faces_as_vertices():
            if len(face) <= max_vertex_count:
                yield face
            else:
                yield from mapbox_earcut_3d(face)

    def mesh_tessellation(self, max_vertex_count: int = 4) -> MeshTransformer:
        """Returns a new :class:`MeshTransformer` instance, where each face has
        no more vertices than the given `max_vertex_count`.

        The `fast` mode uses a shortcut for faces with less than 6 vertices
        which may not work for concave faces!
        """
        mesh = MeshVertexMerger()
        for face in self.tessellation(max_vertex_count=max_vertex_count):
            mesh.add_face(face)
        return MeshTransformer.from_builder(mesh)

    def flip_normals(self) -> None:
        """Flips the normals of all faces by reversing the vertex order inplace."""
        self.faces = list(flip_face_normals(self.faces))

    def separate_meshes(self) -> list[MeshTransformer]:
        """A single :class:`MeshBuilder` instance can store multiple separated
        meshes. This function returns this separated meshes as multiple
        :class:`MeshTransformer` instances.
        """
        return list(separate_meshes(self))

    def normalize_faces(self) -> None:
        """Removes duplicated vertex indices from faces and stores all faces as
        open faces, where the last vertex is not coincident with the first
        vertex.
        """
        self.faces = list(normalize_faces(self.faces, close=False))

    def face_orientation_detector(self, reference: int = 0) -> FaceOrientationDetector:
        """Returns a :class:`FaceOrientationDetector` or short `fod` instance.
        The forward orientation is defined by the `reference` face which is
        0 by default.

        The `fod` can check if all faces are reachable from the reference face
        and if all faces have the same orientation. The `fod` can be reused to
        unify the face orientation of the mesh.

        """
        return FaceOrientationDetector(self, reference=reference)

    def unify_face_normals(
        self, *, fod: Optional[FaceOrientationDetector] = None
    ) -> MeshTransformer:
        """Returns a new :class:`MeshTransformer` object with unified
        face normal vectors of all faces.
        The forward direction (not necessarily outwards) is defined by the
        face-normals of the majority of the faces.
        This function can not process non-manifold meshes (more than two faces
        are connected by a single edge) or multiple disconnected meshes in a
        single :class:`MeshBuilder` object.

        It is possible to pass in an existing :class:`FaceOrientationDetector`
        instance as argument `fod`.

        Raises:
            NonManifoldError: non-manifold mesh
            MultipleMeshesError: the :class:`MeshBuilder` object contains multiple disconnected meshes

        """
        return unify_face_normals_by_majority(self, fod=fod)

    def unify_face_normals_by_reference(
        self,
        reference: int = 0,
        *,
        force_outwards=False,
        fod: Optional[FaceOrientationDetector] = None,
    ) -> MeshTransformer:
        """Returns a new :class:`MeshTransformer` object with unified
        face normal vectors of all faces.
        The forward direction (not necessarily outwards) is defined by the
        reference face, which is the first face of the `mesh` by default.
        This function can not process non-manifold
        meshes (more than two faces are connected by a single edge) or multiple
        disconnected meshes in a single :class:`MeshBuilder` object.

        The outward direction of all face normals can be forced by stetting
        the argument `force_outwards` to ``True`` but this works only for closed
        surfaces, and it's time-consuming!

        It is not possible to check for a closed surface as long the face normal
        vectors are not unified. But it can be done afterward by the attribute
        :meth:`MeshDiagnose.is_closed_surface` to see if the result is
        trustworthy.

        It is possible to pass in an existing :class:`FaceOrientationDetector`
        instance as argument `fod`.

        Args:
            reference: index of the reference face
            force_outwards: forces face-normals to point outwards, this works
                only for closed surfaces, and it's time-consuming!
            fod: :class:`FaceOrientationDetector` instance

        Raises:
            ValueError: non-manifold mesh or the :class:`MeshBuilder` object
                contains multiple disconnected meshes

        """
        mesh = unify_face_normals_by_reference(self, reference=reference, fod=fod)
        if force_outwards:
            _force_face_normals_pointing_outwards(mesh, reference)
        return mesh


class MeshTransformer(MeshBuilder):
    """A mesh builder with inplace transformation support."""

    def transform(self, matrix: Matrix44):
        """Transform mesh inplace by applying the transformation `matrix`.

        Args:
            matrix: 4x4 transformation matrix as :class:`~ezdxf.math.Matrix44`
                object

        """
        self.vertices = list(matrix.transform_vertices(self.vertices))
        return self

    def translate(self, dx: Union[float, UVec] = 0, dy: float = 0, dz: float = 0):
        """Translate mesh inplace.

        Args:
            dx: translation in x-axis or translation vector
            dy: translation in y-axis
            dz: translation in z-axis

        """
        if isinstance(dx, (float, int)):
            t = Vec3(dx, dy, dz)
        else:
            t = Vec3(dx)
        self.vertices = [t + v for v in self.vertices]
        return self

    def scale(self, sx: float = 1, sy: float = 1, sz: float = 1):
        """Scale mesh inplace.

        Args:
            sx: scale factor for x-axis
            sy: scale factor for y-axis
            sz: scale factor for z-axis

        """
        self.vertices = [Vec3(x * sx, y * sy, z * sz) for x, y, z in self.vertices]
        return self

    def scale_uniform(self, s: float):
        """Scale mesh uniform inplace.

        Args:
            s: scale factor for x-, y- and z-axis

        """
        self.vertices = [v * s for v in self.vertices]
        return self

    def rotate_x(self, angle: float):
        """Rotate mesh around x-axis about `angle` inplace.

        Args:
            angle: rotation angle in radians

        """
        self.vertices = list(Matrix44.x_rotate(angle).transform_vertices(self.vertices))
        return self

    def rotate_y(self, angle: float):
        """Rotate mesh around y-axis about `angle` inplace.

        Args:
            angle: rotation angle in radians

        """
        self.vertices = list(Matrix44.y_rotate(angle).transform_vertices(self.vertices))
        return self

    def rotate_z(self, angle: float):
        """Rotate mesh around z-axis about `angle` inplace.

        Args:
            angle: rotation angle in radians

        """
        self.vertices = list(Matrix44.z_rotate(angle).transform_vertices(self.vertices))
        return self

    def rotate_axis(self, axis: UVec, angle: float):
        """Rotate mesh around an arbitrary axis located in the origin (0, 0, 0)
        about `angle`.

        Args:
            axis: rotation axis as Vec3
            angle: rotation angle in radians

        """
        self.vertices = list(
            Matrix44.axis_rotate(axis, angle).transform_vertices(self.vertices)
        )
        return self


def _subdivide(mesh, quads=True) -> MeshVertexMerger:
    """Returns a new :class:`MeshVertexMerger` object with subdivided faces
    and edges.

    Args:
         quads: create quad faces if ``True`` else create triangles

    """
    new_mesh = MeshVertexMerger()
    for vertices in mesh.faces_as_vertices():
        if len(vertices) < 3:
            continue
        for face in subdivide_face(vertices, quads):
            new_mesh.add_face(face)
    return new_mesh


class MeshVertexMerger(MeshBuilder):
    """Subclass of :class:`MeshBuilder`

    Mesh with unique vertices and no doublets, but needs extra memory for
    bookkeeping.

    :class:`MeshVertexMerger` creates a key for every vertex by rounding its
    components by the Python :func:`round` function and a given `precision`
    value. Each vertex with the same key gets the same vertex index, which is
    the index of first vertex with this key, so all vertices with the same key
    will be located at the location of this first vertex. If you want an average
    location of all vertices with the same key use the
    :class:`MeshAverageVertexMerger` class.

    Args:
        precision: floating point precision for vertex rounding

    """

    # can not support vertex transformation
    def __init__(self, precision: int = 6):
        """
        Args:
            precision: floating point precision for vertex rounding

        """
        super().__init__()
        self.ledger: dict[Vec3, int] = {}
        self.precision: int = precision

    def add_vertices(self, vertices: Iterable[UVec]) -> Face:
        """Add new `vertices` only, if no vertex with identical (x, y, z)
        coordinates already exist, else the index of the existing vertex is
        returned as index of the added vertices.

        Args:
            vertices: list of vertices, vertex as (x, y, z) tuple or
                :class:`~ezdxf.math.Vec3` objects

        Returns:
            indices of the added `vertices`

        """
        indices = []
        precision = self.precision
        for vertex in Vec3.generate(vertices):
            key = vertex.round(precision)
            try:
                indices.append(self.ledger[key])
            except KeyError:
                index = len(self.vertices)
                self.vertices.append(vertex)
                self.ledger[key] = index
                indices.append(index)
        return tuple(indices)

    def index(self, vertex: UVec) -> int:
        """Get index of `vertex`, raises :class:`IndexError` if not found.

        Args:
            vertex: ``(x, y, z)`` tuple or :class:`~ezdxf.math.Vec3` object

        (internal API)
        """
        try:
            return self.ledger[Vec3(vertex).round(self.precision)]
        except KeyError:
            raise IndexError(f"Vertex {str(vertex)} not found.")

    @classmethod
    def from_builder(cls, other: MeshBuilder) -> MeshVertexMerger:
        """Create new mesh from other mesh builder."""
        # rebuild from scratch to create a valid ledger
        return cls.from_mesh(other)


class MeshAverageVertexMerger(MeshBuilder):
    """Subclass of :class:`MeshBuilder`

    Mesh with unique vertices and no doublets, but needs extra memory for
    bookkeeping and runtime for calculation of average vertex location.

    :class:`MeshAverageVertexMerger` creates a key for every vertex by rounding
    its components by the Python :func:`round` function and a given `precision`
    value. Each vertex with the same key gets the same vertex index, which is the
    index of first vertex with this key, the difference to the
    :class:`MeshVertexMerger` class is the calculation of the average location
    for all vertices with the same key, this needs extra memory to keep track of
    the count of vertices for each key and extra runtime for updating the vertex
    location each time a vertex with an existing key is added.

    Args:
        precision: floating point precision for vertex rounding

    """

    # can not support vertex transformation
    def __init__(self, precision: int = 6):
        super().__init__()
        self.ledger: dict[
            Vec3, tuple[int, int]
        ] = {}  # each key points to a tuple (vertex index, vertex count)
        self.precision: int = precision

    def add_vertices(self, vertices: Iterable[UVec]) -> Face:
        """Add new `vertices` only, if no vertex with identical ``(x, y, z)``
        coordinates already exist, else the index of the existing vertex is
        returned as index of the added vertices.

        Args:
            vertices: list of vertices, vertex as ``(x, y, z)`` tuple or
            :class:`~ezdxf.math.Vec3` objects

        Returns:
            tuple: indices of the `vertices` added to the
            :attr:`~MeshBuilder.vertices` list

        """
        indices = []
        precision = self.precision
        for vertex in Vec3.generate(vertices):
            key = vertex.round(precision)
            try:
                index, count = self.ledger[key]
            except KeyError:  # new key
                index = len(self.vertices)
                self.vertices.append(vertex)
                self.ledger[key] = (index, 1)
            else:  # update key entry
                # calculate new average location
                average = (self.vertices[index] * count) + vertex
                count += 1
                # update vertex location
                self.vertices[index] = average / count
                # update ledger
                self.ledger[key] = (index, count)
            indices.append(index)
        return tuple(indices)

    def index(self, vertex: UVec) -> int:
        """Get index of `vertex`, raises :class:`IndexError` if not found.

        Args:
            vertex: ``(x, y, z)`` tuple or :class:`~ezdxf.math.Vec3` object

        (internal API)
        """
        try:
            return self.ledger[Vec3(vertex).round(self.precision)][0]
        except KeyError:
            raise IndexError(f"Vertex {str(vertex)} not found.")

    @classmethod
    def from_builder(cls, other: MeshBuilder) -> MeshAverageVertexMerger:
        """Create new mesh from other mesh builder."""
        # rebuild from scratch to create a valid ledger
        return cls.from_mesh(other)


class _XFace:
    __slots__ = ("fingerprint", "indices", "_orientation")

    def __init__(self, indices: Face):
        self.fingerprint: int = hash(indices)
        self.indices: Face = indices
        self._orientation: Vec3 = VEC3_SENTINEL

    def orientation(self, vertices: Sequence[Vec3], precision: int = 4) -> Vec3:
        if self._orientation is VEC3_SENTINEL:
            orientation = NULLVEC
            v0, v1, *v = [vertices[i] for i in self.indices]
            for v2 in v:
                try:
                    orientation = normal_vector_3p(v0, v1, v2).round(precision)
                    break
                except ZeroDivisionError:
                    continue
            self._orientation = orientation
        return self._orientation


def _merge_adjacent_coplanar_faces(
    vertices: list[Vec3], faces: list[Face], precision: int = 4
) -> MeshVertexMerger:
    oriented_faces: dict[Vec3, list[_XFace]] = {}
    extended_faces: list[_XFace] = []
    for face in faces:
        if len(face) < 3:
            raise ValueError("found invalid face count < 3")
        xface = _XFace(face)
        extended_faces.append(xface)
        oriented_faces.setdefault(xface.orientation(vertices, precision), []).append(
            xface
        )

    mesh = MeshVertexMerger()
    done = set()
    for xface in extended_faces:
        if xface.fingerprint in done:
            continue
        done.add(xface.fingerprint)
        face = xface.indices
        orientation = xface.orientation(vertices, precision)
        parallel_faces = oriented_faces[orientation]
        face_set = set(face)
        for parallel_face in parallel_faces:
            if parallel_face.fingerprint in done:
                continue
            common_vertices = face_set.intersection(set(parallel_face.indices))
            # connection by at least 2 vertices required:
            if len(common_vertices) > 1:
                if len(common_vertices) == len(parallel_face.indices):
                    face = merge_full_patch(face, parallel_face.indices)
                else:
                    try:
                        face = merge_connected_paths(face, parallel_face.indices)
                    except (NodeMergingError, DegeneratedPathError):
                        continue
                done.add(parallel_face.fingerprint)
                face_set = set(face)
        v0 = list(remove_colinear_face_vertices([vertices[i] for i in face]))
        mesh.add_face(v0)
    return mesh


VEC3_SENTINEL = Vec3(0, 0, 0)


def remove_colinear_face_vertices(vertices: Sequence[Vec3]) -> Iterator[Vec3]:
    def get_direction(v1: Vec3, v2: Vec3):
        return (v2 - v1).normalize()

    if len(vertices) < 3:
        yield from vertices
        return

    # remove duplicated vertices
    _vertices: list[Vec3] = [vertices[0]]
    for v in vertices[1:]:
        if not v.isclose(_vertices[-1]):
            _vertices.append(v)

    if len(_vertices) < 3:
        if len(_vertices) == 1:
            _vertices.append(_vertices[0])
        yield from _vertices
        return

    start = _vertices[0]
    prev_vertex = VEC3_SENTINEL
    current_direction = VEC3_SENTINEL
    start_index = 0

    # find start direction
    yield start
    while current_direction is VEC3_SENTINEL:
        start_index += 1
        try:
            prev_vertex = vertices[start_index]
        except IndexError:
            yield prev_vertex
            return
        current_direction = get_direction(start, prev_vertex)

    yielded_anything = False
    _vertices.append(start)
    for vertex in _vertices[start_index:]:
        try:
            if get_direction(start, vertex).isclose(current_direction):
                prev_vertex = vertex
                continue
        except ZeroDivisionError:
            continue
        yield prev_vertex
        yielded_anything = True
        start = prev_vertex
        current_direction = get_direction(start, vertex)
        prev_vertex = vertex

    if not yielded_anything:
        yield _vertices[-2]  # last vertex


def merge_connected_paths(p1: Sequence[int], p2: Sequence[int]) -> Sequence[int]:
    def build_nodes(p: Sequence[int]):
        nodes = {e1: e2 for e1, e2 in zip(p, p[1:])}
        nodes[p[-1]] = p[0]
        return nodes

    current_path = build_nodes(p1)
    other_path = build_nodes(p2)
    current_node = p1[0]
    finish = p1[0]
    connected_path = [current_node]
    while True:
        try:
            next_node = current_path[current_node]
        except KeyError:
            raise NodeMergingError
        if next_node in other_path:
            current_path, other_path = other_path, current_path
        if next_node == finish:
            break
        current_node = next_node
        if current_node in connected_path:
            # node duplication is an error, e.g. two path are only connected
            # by one node:
            raise NodeMergingError
        connected_path.append(current_node)

    if len(connected_path) < 3:
        raise DegeneratedPathError
    return connected_path


def merge_full_patch(path: Sequence[int], patch: Sequence[int]):
    count = len(path)
    new_path = []
    for pos, node in enumerate(path):
        prev = path[pos - 1]
        succ = path[(pos + 1) % count]
        if prev in patch and succ in patch:
            continue
        new_path.append(node)
    return new_path


class Lump:
    def __init__(self, face: Face):
        self.edges: set[Edge] = set()
        self.faces: list[Face] = [face]
        for a, b in face_edges(face):
            # sort vertex indices to guarantee: edge a,b == edge b,a
            self.edges.add((a, b) if a <= b else (b, a))

    def is_connected(self, other: Lump) -> bool:
        return not self.edges.isdisjoint(other.edges)

    def merge(self, other: Lump):
        self.faces.extend(other.faces)
        self.edges.update(other.edges)


def merge_lumps(lumps: Iterable[Lump]) -> list[Lump]:
    merged_lumps = _merge_lumps(lumps)
    prior_len = 0
    while 1 < len(merged_lumps) != prior_len:
        prior_len = len(merged_lumps)
        merged_lumps = _merge_lumps(merged_lumps)
    return merged_lumps


def _merge_lumps(lumps: Iterable[Lump]) -> list[Lump]:
    merged_lumps: list[Lump] = []
    for lump in lumps:
        for base in merged_lumps:
            if lump.is_connected(base):
                base.merge(lump)
                break
        else:
            merged_lumps.append(lump)
    return merged_lumps


def separate_meshes(m: MeshBuilder) -> Iterator[MeshTransformer]:
    """Returns the separated meshes in a single :class:`MeshBuilder` instance
    as multiple :class:`MeshTransformer` instances.
    """
    # At the beginning each face is a separated lump and all connected faces
    # should be merged:
    disconnected_lumps = list(merge_lumps(Lump(face) for face in open_faces(m.faces)))
    if len(disconnected_lumps) > 1:
        vertices = m.vertices
        # create new separated meshes:
        for lump in disconnected_lumps:
            mesh = MeshVertexMerger()
            for face in lump.faces:
                mesh.add_face(vertices[i] for i in face)
            yield MeshTransformer.from_builder(mesh)
    else:
        yield MeshTransformer.from_builder(m)


def have_away_pointing_normals(f0: Sequence[Vec3], f1: Sequence[Vec3]) -> bool:
    """Returns ``True`` if the normals of the two faces are pointing
    away from the center of the two faces.

    .. warning::

        This does not work for any arbitrary face pair!

    """
    c0 = Vec3.sum(f0) / len(f0)
    c1 = Vec3.sum(f1) / len(f1)
    n0 = normal_vector_3p(f0[0], f0[1], f0[2]) * 0.5
    n1 = normal_vector_3p(f1[0], f1[1], f1[2]) * 0.5
    e0 = c0 + n0
    e1 = c1 + n1
    # distance of centroids
    d0 = c0.distance(c1)
    # distance of normal end points
    d1 = e0.distance(e1)
    return d1 > d0


def face_normals_after_transformation(m: Matrix44) -> bool:
    """Returns the state of face-normals of the unit-cube after the
    transformation `m` was applied.

        - ``True``: face-normals pointing outwards
        - ``False``: face-normals pointing inwards

    The state before the transformation is outward-pointing face-normals.

    """
    from .forms import cube

    unit_cube = cube(True).transform(m)
    bottom, _, _, _, _, top = unit_cube.faces_as_vertices()
    # it is not necessary to check all 3 axis!
    return have_away_pointing_normals(bottom, top)


def _make_edge_mapping(faces: Iterable[Face]) -> dict[Edge, list[Face]]:
    mapping: dict[Edge, list[Face]] = {}
    for face in faces:
        for edge in face_edges(face):
            mapping.setdefault(edge, []).append(face)
    return mapping


class FaceOrientationDetector:
    """
    Helper class for face orientation and face normal vector detection. Use the
    method :meth:`MeshBuilder.face_orientation_detector` to create an instance.

    The face orientation detector classifies the faces of a mesh by their
    forward or backward orientation.
    The forward orientation is defined by a reference face, which is the first
    face of the mesh by default and this orientation is not necessarily outwards.

    This class has some overlapping features with :class:`MeshDiagnose` but it
    has a longer setup time and needs more memory than :class:`MeshDiagnose`.

    Args:
        mesh: source mesh as :class:`MeshBuilder` object
        reference: index of the reference face

    """

    def __init__(self, mesh: MeshBuilder, reference: int = 0):
        self._mesh = mesh
        self.edge_mapping: dict[Edge, list[Face]] = _make_edge_mapping(mesh.faces)
        self.reference = reference
        self.is_manifold = True  # 2-manifold is meant
        self.forward: dict[int, Face] = dict()
        self.backward: dict[int, Face] = dict()
        self.classify_faces(reference)

    @property
    def has_uniform_face_normals(self) -> bool:
        """Returns ``True`` if all reachable faces are forward oriented
        according to the reference face.
        """
        return len(self.backward) == 0

    @property
    def all_reachable(self) -> bool:
        """Returns ``True`` if all faces are reachable from the reference face
        same as property :attr:`is_single_mesh`.
        """
        return len(self._mesh.faces) == sum(self.count)

    @property
    def is_single_mesh(self) -> bool:
        """Returns ``True`` if only a single mesh is present same as property
        :attr:`all_reachable`.
        """
        return self.all_reachable

    @property
    def count(self) -> tuple[int, int]:
        """Returns the count of forward and backward oriented faces."""
        return len(self.forward), len(self.backward)

    @property
    def forward_faces(self) -> Iterator[Face]:
        """Yields all forward oriented faces."""
        return iter(self.forward.values())

    @property
    def backward_faces(self) -> Iterator[Face]:
        """Yields all backward oriented faces."""
        return iter(self.backward.values())

    def classify_faces(self, reference: int = 0) -> None:
        """Detect the forward and backward oriented faces.

        The forward and backward orientation has to be defined by a `reference`
        face.
        """

        def add_forward(f: Face):
            forward[id(f)] = f

        def add_backward(f: Face):
            backward[id(f)] = f

        def add_face_to_process(f: Face, orientation: bool):
            key = id(f)
            if key not in forward and key not in backward:  # unprocessed face!
                process_faces.append((f, orientation))

        def add_adjacent_faces_to_process_queue(f: Face, orientation: bool):
            for edge in face_edges(f):
                # find adjacent faces at this edge with same edge orientation:
                linked_faces = edge_mapping[edge]
                if len(linked_faces) > 1:
                    # these faces are backward oriented faces:
                    for linked_face in linked_faces:
                        if linked_face is f:  # skip legit current face
                            continue
                        add_face_to_process(linked_face, not orientation)

                # find all adjacent faces at this edge with reversed edges:
                try:
                    linked_faces = edge_mapping[edge[1], edge[0]]
                except KeyError:
                    # open surface or backward oriented faces present
                    continue
                if len(linked_faces) == 1:
                    add_face_to_process(linked_faces[0], orientation)
                else:  # non-manifold mesh
                    # none of the linked face is processed!
                    self.is_manifold = False

        self.reference = int(reference)
        self.is_manifold = True
        edge_mapping = self.edge_mapping
        forward: dict[int, Face] = dict()
        backward: dict[int, Face] = dict()
        # the reference face defines the forward orientation
        process_faces: list[tuple[Face, bool]] = [(self._mesh.faces[reference], True)]
        while len(process_faces):
            # current face and orientation, True = forward, False = backward
            face, face_orientation = process_faces.pop(0)
            if face_orientation:
                add_forward(face)
            else:
                add_backward(face)
            add_adjacent_faces_to_process_queue(face, face_orientation)

        self.forward = forward
        self.backward = backward

    @property
    def is_closed_surface(self) -> bool:
        """Returns ``True`` if the mesh has a closed surface.
        This method does not require a unified face orientation.
        If multiple separated meshes are present the state is only ``True`` if
        **all** meshes have a closed surface.

        Returns ``False`` for non-manifold meshes.

        """
        if not self.is_manifold:
            return False
        empty: list[Face] = []
        edge_mapping = self.edge_mapping
        # For a closed surface all edges have to connect exact 2 faces.
        for edge in edge_mapping.keys():
            forward_connected_faces = edge_mapping.get(edge, empty)
            reversed_edge = edge[1], edge[0]
            backward_connected_faces = edge_mapping.get(reversed_edge, empty)
            if len(forward_connected_faces) + len(backward_connected_faces) != 2:
                return False
        return True

    def is_reference_face_pointing_outwards(self) -> bool:
        """Returns ``True`` if the normal vector of the reference face is
        pointing outwards. This works only for meshes with unified faces which
        represent a closed surfaces, and it's a time-consuming calculation!

        """
        from ezdxf.math import is_face_normal_pointing_outwards

        faces = list(self._mesh.faces_as_vertices())
        reference_face = faces[self.reference]
        return is_face_normal_pointing_outwards(faces, reference_face)


def unify_face_normals_by_reference(
    mesh: MeshBuilder,
    *,
    fod: Optional[FaceOrientationDetector] = None,
    reference: int = 0,
) -> MeshTransformer:
    """Unifies the orientation of all faces of a :class:`MeshBuilder` object.
    The forward orientation is defined by a reference face, which is the first
    face of the `mesh` by default. This function can not process non-manifold
    meshes (more than two faces are connected by a single edge) or multiple
    disconnected meshes in a single :class:`MeshBuilder` object.
    Returns always a copy of the source `mesh` as :class:`MeshTransformer`
    object.

    Args:
        mesh: source :class:`MeshBuilder` object
        fod: an already created :class:`FaceOrientationDetector`
            instance or ``None`` to create one internally.
        reference: index of the reference face for the internally created
            :class:`FaceOrientationDetector` instance

    Raises:
        NonManifoldError: non-manifold mesh
        MultipleMeshesError: :class:`MeshBuilder` object contains multiple
            disconnected meshes

    """
    if fod is None:
        fod = FaceOrientationDetector(mesh, reference)

    def backward_selector(detector: FaceOrientationDetector):
        return detector.backward

    return _unify_face_normals(mesh, fod, backward_selector)


def unify_face_normals_by_majority(
    mesh: MeshBuilder,
    *,
    fod: Optional[FaceOrientationDetector] = None,
) -> MeshTransformer:
    """Unifies the orientation of all faces of a :class:`MeshBuilder` object.
    The forward orientation is defined by the orientation of the majority of
    the faces.
    This function can not process non-manifold meshes (more than two faces are
    connected by a single edge) or multiple disconnected meshes in a single
    :class:`MeshBuilder` object.
    Returns always a copy of the source `mesh` as :class:`MeshTransformer`
    object.

    Args:
        mesh: source :class:`MeshBuilder` object
        fod: an already created :class:`FaceOrientationDetector`
            instance or ``None`` to create one internally.

    Raises:
        NonManifoldError: non-manifold mesh
        MultipleMeshesError: :class:`MeshBuilder` object contains multiple
            disconnected meshes

    """
    if fod is None:
        fod = FaceOrientationDetector(mesh)

    def backward_selector(detector: FaceOrientationDetector):
        count0, count1 = detector.count
        return detector.backward if count0 >= count1 else detector.forward

    return _unify_face_normals(mesh, fod, backward_selector)


def _unify_face_normals(
    mesh: MeshBuilder,
    fod: FaceOrientationDetector,
    backward_selector,
) -> MeshTransformer:
    """Unifies all faces of a MeshBuilder. The backward_selector() function
    returns the face ledger which represents the backward oriented faces.
    """
    if not fod.is_manifold:
        raise NonManifoldMeshError("non-manifold mesh")
    if not fod.all_reachable:
        raise MultipleMeshesError(
            f"not all faces are reachable from reference face #{fod.reference}"
        )

    new_mesh = MeshTransformer()
    new_mesh.vertices = list(mesh.vertices)
    if not fod.has_uniform_face_normals:
        backward = backward_selector(fod)
        faces = []
        for face in mesh.faces:
            if id(face) in backward:
                faces.append(tuple(reversed(face)))
            else:
                faces.append(tuple(face))
        new_mesh.faces = faces  # type: ignore
    else:
        new_mesh.faces = list(mesh.faces)
    return new_mesh


def _force_face_normals_pointing_outwards(
    mesh: MeshBuilder, reference: int = 0
) -> None:
    """Detect the orientation of the reference face and flip all face normals
    if required.

    Requirements: all face normals have to be unified and the mesh has a closed
    surface

    """
    from ezdxf.math import is_face_normal_pointing_outwards

    faces = list(mesh.faces_as_vertices())
    reference_face = faces[reference]
    if not is_face_normal_pointing_outwards(faces, reference_face):
        mesh.flip_normals()


def normalize_mesh(mesh: MeshBuilder) -> MeshTransformer:
    """Returns a new mesh with these properties:

        - one mesh: removes all meshes except the first mesh starting at the
          first face
        - optimized vertices: merges shared vertices
        - normalized faces: removes duplicated face vertices and faces with less
          than 3 vertices
        - open faces: all faces are open faces where first vertex != last vertex
        - unified face-orientation: all faces have the same orientation (ccw or cw)
        - ccw face-orientation: all face-normals are pointing outwards if the
          mesh has a closed surface

    Raises:
        NonManifoldMeshError: non-manifold mesh

    """
    new_mesh = mesh.optimize_vertices()
    new_mesh.normalize_faces()
    fod = new_mesh.face_orientation_detector(reference=0)

    if not fod.is_manifold:
        raise NonManifoldMeshError("non-manifold mesh")

    if not fod.all_reachable:  # extract first mesh
        new_mesh = list(separate_meshes(new_mesh))[0]
        fod = new_mesh.face_orientation_detector(reference=0)

    if not fod.has_uniform_face_normals:
        new_mesh = unify_face_normals_by_reference(new_mesh, fod=fod)
        fod = new_mesh.face_orientation_detector()

    if fod.is_closed_surface and not fod.is_reference_face_pointing_outwards():
        new_mesh.flip_normals()
    return new_mesh



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf
from ezdxf.addons.drawing.properties import RenderContext, is_valid_color
from ezdxf.addons.acadctb import ColorDependentPlotStyles
from ezdxf import colors
from ezdxf.entities import Layer, factory
from ezdxf.lldxf import const


@pytest.fixture(scope="module")
def doc():
    d = ezdxf.new(setup=True)
    d.layers.add("Test", color=5, linetype="DOT", lineweight=70)
    msp = d.modelspace()
    msp.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 256,  # by layer
            "linetype": "BYLAYER",
            "lineweight": -1,  # by layer
            "layer": "Test",
        },
    )
    msp.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 1,  # red: ff0000
            "linetype": "DASHED",
            "lineweight": 50,  # 0.50
            "layer": "Test",
        },
    )
    blk = d.blocks.new("MyBlock")
    blk.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 0,  # by block
            "linetype": "BYBLOCK",
            "lineweight": -2,  # by block
            "layer": "Test",
        },
    )
    blk.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 1,  # red: ff0000
            "linetype": "DASHED",
            "lineweight": 50,  # 0.50
            "layer": "Test",
        },
    )
    blk.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 256,  # by layer
            "linetype": "BYLAYER",
            "lineweight": -1,  # by layer
            "layer": "Test",
        },
    )
    msp.add_blockref(
        "MyBlock",
        insert=(0, 0),
        dxfattribs={
            "color": 3,  # green: 00ff00
            "linetype": "CENTER",
            "lineweight": 13,  # 0.13
        },
    )
    return d


def test_load_default_ctb(doc):
    ctx = RenderContext(doc, ctb="color.ctb")
    assert bool(ctx.plot_styles) is True
    assert ctx.plot_styles[1].color == (255, 0, 0)


def test_set_color_dependent_plot_style_as_object_instance(doc):
    ctb = ColorDependentPlotStyles(description="my_style")
    ctb.new_style(1).color = (22, 33, 44)
    ctx = RenderContext(doc, ctb=ctb)
    assert bool(ctx.plot_styles) is True
    assert ctx.plot_styles[1].color == (22, 33, 44)


def test_new_ctb(doc):
    ctx = RenderContext(doc)
    assert bool(ctx.plot_styles) is True
    assert ctx.plot_styles[1].color == (255, 0, 0)


def test_resolve_entity_visibility():
    doc = ezdxf.new()
    layout = doc.modelspace()
    doc.layers.new(name="visible", dxfattribs={"color": 0})
    doc.layers.new(name="invisible", dxfattribs={"color": -1})  # color < 0 => invisible
    doc.layers.new(name="frozen", dxfattribs={"flags": Layer.FROZEN})  # also invisible
    doc.layers.new(
        name="noplot", dxfattribs={"plot": 0}
    )  # visible in the CAD application but not when exported

    for export_mode in (False, True):
        ctx = RenderContext(layout.doc, export_mode=export_mode)

        text = layout.add_text(
            "a", dxfattribs={"invisible": 0, "layer": "non_existent"}
        )
        assert ctx.resolve_visible(text) is True

        text = layout.add_text("a", dxfattribs={"invisible": 0, "layer": "visible"})
        assert ctx.resolve_visible(text) is True

        for layer in ["invisible", "frozen"]:
            text = layout.add_text("a", dxfattribs={"invisible": 0, "layer": layer})
            assert ctx.resolve_visible(text) is False

        for layer in [
            "non_existent",
            "visible",
            "invisible",
            "frozen",
            "noplot",
        ]:
            text = layout.add_text("a", dxfattribs={"invisible": 1, "layer": layer})
            assert ctx.resolve_visible(text) is False

    ctx = RenderContext(layout.doc, export_mode=False)
    text = layout.add_text("a", dxfattribs={"invisible": 0, "layer": "noplot"})
    assert ctx.resolve_visible(text) is True

    ctx = RenderContext(layout.doc, export_mode=True)
    text = layout.add_text("a", dxfattribs={"invisible": 0, "layer": "noplot"})
    assert ctx.resolve_visible(text) is False


def test_resolve_attrib_visibility():
    doc = ezdxf.new()
    layout = doc.modelspace()
    block = doc.blocks.new(name="block")
    doc.layers.new(name="invisible", dxfattribs={"color": -1})  # color < 0 => invisible

    block.add_attdef("att1", (0, 0), "", dxfattribs={})
    block.add_attdef("att2", (0, 0), "", dxfattribs={"flags": const.ATTRIB_INVISIBLE})
    block.add_attdef("att3", (0, 0), "", dxfattribs={"layer": "invisible"})

    i = layout.add_blockref("block", (0, 0))
    i.add_auto_attribs({"att1": "abc", "att2": "def", "att3": "hij"})

    assert not i.attribs[0].is_invisible
    assert i.attribs[1].is_invisible
    assert not i.attribs[2].is_invisible

    ctx = RenderContext(layout.doc)
    assert ctx.resolve_visible(i.attribs[0]) is True
    assert ctx.resolve_visible(i.attribs[1]) is False
    assert ctx.resolve_visible(i.attribs[2]) is False


def test_resolve_entity_color(doc):
    ctx = RenderContext(doc)
    msp = doc.modelspace()
    lines = msp.query("LINE")
    line1 = ctx.resolve_all(lines[0])
    assert line1.color == "#0000ff"  # by layer

    line2 = ctx.resolve_all(lines[1])
    assert line2.color == "#ff0000"


def test_existing_true_color_overrides_any_aci_color(doc):
    ctx = RenderContext(doc)
    line = factory.new("LINE")
    line.rgb = (255, 1, 1)
    for color in range(const.BYLAYER + 1):
        line.dxf.color = color
        props = ctx.resolve_all(line)
        assert props.color == "#ff0101", "true_color should override any ACI color"


def test_resolve_entity_linetype(doc):
    ctx = RenderContext(doc)
    msp = doc.modelspace()
    lines = msp.query("LINE")

    line1 = ctx.resolve_all(lines[0])
    assert line1.linetype_name == "DOT"  # by layer

    line2 = ctx.resolve_all(lines[1])
    assert line2.linetype_name == "DASHED"


def test_resolve_entity_lineweight(doc):
    ctx = RenderContext(doc)
    msp = doc.modelspace()
    lines = msp.query("LINE")

    line1 = ctx.resolve_all(lines[0])
    assert line1.lineweight == 0.70  # by layer

    line2 = ctx.resolve_all(lines[1])
    assert line2.lineweight == 0.50


def test_resolve_block_entities(doc):
    ctx = RenderContext(doc)
    msp = doc.modelspace()
    blockref = msp.query("INSERT").first
    ctx.push_state(ctx.resolve_all(blockref))
    assert ctx.inside_block_reference is True
    lines = list(blockref.virtual_entities())

    # properties by block
    line1 = ctx.resolve_all(lines[0])
    assert lines[0].dxf.linetype == "BYBLOCK"
    assert line1.color == "#00ff00"
    assert line1.linetype_name == "CENTER"
    assert line1.lineweight == 0.13

    # explicit properties
    line2 = ctx.resolve_all(lines[1])
    assert lines[1].dxf.linetype == "DASHED"
    assert line2.color == "#ff0000"
    assert line2.linetype_name == "DASHED"
    assert line2.lineweight == 0.50

    # properties by layer 'Test'
    line3 = ctx.resolve_all(lines[2])
    assert lines[2].dxf.linetype == "BYLAYER"
    assert line3.color == "#0000ff"
    assert line3.linetype_name == "DOT"
    assert line3.lineweight == 0.70

    ctx.pop_state()
    assert ctx.inside_block_reference is False


class TestResolveLayerACIColor7:
    @pytest.fixture
    def entity(self):
        # Default layer is '0' with default ACI color 7
        # Default entity color is BYLAYER
        return factory.new("LINE")

    @pytest.fixture(scope="class")
    def ctx(self):
        doc = ezdxf.new()
        doc.layers.new(
            "TrueColor",
            dxfattribs={"true_color": ezdxf.rgb2int((0xB0, 0xB0, 0xB0))},
        )
        context = RenderContext(doc)
        context.set_current_layout(doc.modelspace())
        return context

    def test_dark_background(self, ctx, entity):
        ctx.current_layout_properties.set_colors(bg="#000000")
        assert ctx.resolve_color(entity).upper() == "#FFFFFF"

    def test_light_background(self, ctx, entity):
        ctx.current_layout_properties.set_colors(bg="#FFFFFF")
        assert ctx.resolve_color(entity) == "#000000"

    def test_switch_layout_colors(self, ctx, entity):
        ctx.current_layout_properties.set_colors(bg="#FFFFFF", fg="#A0A0A0")
        assert ctx.resolve_color(entity) == "#A0A0A0"
        ctx.current_layout_properties.set_colors(bg="#EEEEEE", fg="#010101")
        assert ctx.resolve_color(entity) == "#010101"

    def test_color_from_true_color_layer(self, ctx, entity):
        entity.dxf.layer = "TrueColor"
        assert ctx.resolve_color(entity).upper() == "#B0B0B0"

        # Entity ACI color is BYLAYER by default:
        assert entity.dxf.color == const.BYLAYER

        ctx.current_layout_properties.set_colors(bg="#EEEEEE", fg="#010101")
        # But has no meaning if true color is set:
        assert ctx.resolve_color(entity).upper() == "#B0B0B0"


@pytest.mark.parametrize(
    "color, result",
    [
        ("#012345", True),
        ("#456789", True),
        ("#ABCDEF", True),
        ("#abcdef", True),
        ("#ghijkl", False),
        ("000000", False),
        ("ABCDEF", False),
    ],
)
def test_is_valid_color(color, result):
    assert is_valid_color(color) is result


@pytest.mark.parametrize("color", [0, 1.0, (1, 2, 3)])
def test_invalid_color_value_type(color):
    with pytest.raises(TypeError):
        is_valid_color(color)


@pytest.mark.parametrize(
    "color, result",
    [
        ("#00000000", True),
        ("#000000FF", True),
        ("#000000ff", True),
        ("#000000gh", False),
    ],
)
def test_is_valid_transparent_color(color, result):
    assert is_valid_color(color) is result


@pytest.mark.parametrize(
    "color, result",
    [
        ("", False),
        ("#", False),
        ("#0", False),
        ("#00", False),
        ("#000", False),
        ("#0000", False),
        ("#00000", False),
        ("#000000", True),  # RGB color format
        ("#0000000", False),
        ("#00000000", True),  # RGB color format including alpha transparency
        ("#000000000", False),
        ("  #0000", False),
        ("  #000000", False),
    ],
)
def test_is_valid_color_value_length(color, result):
    assert is_valid_color(color) is result


def test_resolve_transparency_from_layer():
    doc = ezdxf.new()
    doc.layers.add("Layer_T50", color=1, transparency=0.50)
    msp = doc.modelspace()
    line = msp.add_line(
        (0, 0),
        (1, 0),
        dxfattribs={
            "color": 5,
            "layer": "Layer_T50",
        },
    )
    context = RenderContext(doc)
    context.set_current_layout(msp)
    prop = context.resolve_all(line)
    assert prop.color == "#0000ff7f"


def test_bylayer_linetype_properties():
    doc = ezdxf.new()
    doc.linetypes.add(
        name="CUSTOM_DASH",
        pattern=[1, 12.0, -3.0],
    )
    doc.layers.add("CustomLayer", linetype="CUSTOM_DASH")
    msp = doc.modelspace()
    line = msp.add_line((0, 1), (1, 0), dxfattribs={"layer": "CustomLayer"})
    ctx = RenderContext(doc)
    properties = ctx.resolve_all(line)

    assert properties.linetype_name == "CUSTOM_DASH"
    assert properties.linetype_pattern == (12.0, 3.0)


if __name__ == "__main__":
    pytest.main([__file__])
